#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

import os
from collections import OrderedDict
from utilsbase import *
from cache import *

VXLAN_UDP_PORT = 4789

class iproute2(utilsBase):
    """ This class contains helper methods to cache and interact with the
    commands in the iproute2 package """

    _cache_fill_done = False
    ipbatchbuf = ''
    ipbatch = False
    ipbatch_pause = False

    def __init__(self, *args, **kargs):
        utilsBase.__init__(self, *args, **kargs)
        if self.CACHE and not iproute2._cache_fill_done:
            self._link_fill()
            self._addr_fill()
            iproute2._cache_fill_done = True
        
    def _link_fill(self, ifacename=None, refresh=False):
        """ fills cache with link information
       
        if ifacename argument given, fill cache for ifacename, else
        fill cache for all interfaces in the system
        """

        linkout = {}
        if iproute2._cache_fill_done and not refresh: return
        try:
            # if ifacename already present, return
            if (ifacename and not refresh and
                    linkCache.get_attr([ifacename, 'ifflag'])):
                return
        except:
            pass
        cmdout = self.link_show(ifacename=ifacename)
        if not cmdout:
            return
        for c in cmdout.splitlines():
            citems = c.split()
            ifnamenlink = citems[1].split('@')
            if len(ifnamenlink) > 1:
                ifname = ifnamenlink[0]
                iflink = ifnamenlink[1].strip(':')
            else:
                ifname = ifnamenlink[0].strip(':')
                iflink = None
            linkattrs = {}
            linkattrs['link'] = iflink
            linkattrs['ifindex'] = citems[0].strip(':')
            flags = citems[2].strip('<>').split(',')
            linkattrs['flags'] = flags
            linkattrs['ifflag'] = 'UP' if 'UP' in flags else 'DOWN'
            for i in range(0, len(citems)):
                if citems[i] == 'mtu': linkattrs['mtu'] = citems[i+1]
                elif citems[i] == 'state': linkattrs['state'] = citems[i+1]
                elif citems[i] == 'link/ether': linkattrs['hwaddress'] = citems[i+1]
                elif citems[i] == 'vlan' and citems[i+1] == 'id':
                    linkattrs['linkinfo'] = {'vlanid' : citems[i+2]}
                elif citems[i] == 'vxlan' and citems[i+1] == 'id':
                    vattrs = {'vxlanid' : citems[i+2],
                              'svcnode' : [],
                              'remote'  : [],
                              'ageing' : citems[i+2],
                              'learning': 'on'}
                    for j in range(i+2, len(citems)):
                        if citems[j] == 'local':
                            vattrs['local'] = citems[j+1]
                        elif citems[j] == 'svcnode':
                            vattrs['svcnode'].append(citems[j+1])
                        elif citems[j] == 'peernode':
                            vattrs['peernode'].append(citems[j+1])
                        elif citems[j] == 'nolearning':
                            vattrs['learning'] = 'off'
                    # get vxlan peer nodes
                    peers = self.get_vxlan_peers(ifname)
                    if peers:
                        vattrs['remote'] = peers
                    linkattrs['linkinfo'] = vattrs
                    break
            #linkattrs['alias'] = self.read_file_oneline(
            #            '/sys/class/net/%s/ifalias' %ifname)
            linkout[ifname] = linkattrs
        [linkCache.update_attrdict([ifname], linkattrs)
                    for ifname, linkattrs in linkout.items()]

    def _addr_filter(self, addr, scope=None):
        default_addrs = ['127.0.0.1/8', '::1/128' , '0.0.0.0']
        if addr in default_addrs:
            return True
        if scope and scope == 'link':
            return True
        return False

    def _addr_fill(self, ifacename=None, refresh=False):
        """ fills cache with address information
       
        if ifacename argument given, fill cache for ifacename, else
        fill cache for all interfaces in the system
        """

        linkout = {}
        if iproute2._cache_fill_done: return
        try:
            # Check if ifacename is already full, in which case, return
            if ifacename:
                linkCache.get_attr([ifacename, 'addrs']) 
                return
        except:
            pass
        cmdout = self.addr_show(ifacename=ifacename)
        if not cmdout:
            return
        for c in cmdout.splitlines():
            citems = c.split()
            ifnamenlink = citems[1].split('@')
            if len(ifnamenlink) > 1:
                ifname = ifnamenlink[0]
            else:
                ifname = ifnamenlink[0].strip(':')
            if not linkout.get(ifname):
                linkattrs = {}
                linkattrs['addrs'] = OrderedDict({})
                try:
                    linkout[ifname].update(linkattrs)
                except KeyError:
                    linkout[ifname] = linkattrs
            if citems[2] == 'inet':
                if self._addr_filter(citems[3], scope=citems[5]):
                    continue
                addrattrs = {}
                addrattrs['scope'] = citems[5]
                addrattrs['type'] = 'inet'
                linkout[ifname]['addrs'][citems[3]] = addrattrs
            elif citems[2] == 'inet6':
                if self._addr_filter(citems[3], scope=citems[5]):
                    continue
                if citems[5] == 'link': continue #skip 'link' addresses
                addrattrs = {}
                addrattrs['scope'] = citems[5]
                addrattrs['type'] = 'inet6'
                linkout[ifname]['addrs'][citems[3]] = addrattrs
        [linkCache.update_attrdict([ifname], linkattrs)
                    for ifname, linkattrs in linkout.items()]

    def _cache_get(self, type, attrlist, refresh=False):
        try:
            if self.DRYRUN:
                return False
            if self.CACHE:
                if not iproute2._cache_fill_done: 
                    self._link_fill()
                    self._addr_fill()
                    iproute2._cache_fill_done = True
                    return linkCache.get_attr(attrlist)
                if not refresh:
                    return linkCache.get_attr(attrlist)
            if type == 'link':
                self._link_fill(attrlist[0], refresh)
            elif type == 'addr':
                self._addr_fill(attrlist[0], refresh)
            else:
                self._link_fill(attrlist[0], refresh)
                self._addr_fill(attrlist[0], refresh)
            return linkCache.get_attr(attrlist)
        except Exception as e:
            self.logger.debug('_cache_get(%s) : [%s]'
                    %(str(attrlist), str(e)))
            pass
        return None

    def _cache_check(self, type, attrlist, value, refresh=False):
        try:
            attrvalue = self._cache_get(type, attrlist, refresh)
            if attrvalue and attrvalue == value:
                return True
        except Exception as e:
            self.logger.debug('_cache_check(%s) : [%s]'
                    %(str(attrlist), str(e)))
            pass
        return False

    def _cache_update(self, attrlist, value):
        if self.DRYRUN: return
        try:
            linkCache.add_attr(attrlist, value)
        except:
            pass

    def _cache_delete(self, attrlist):
        if self.DRYRUN: return
        try:
            linkCache.del_attr(attrlist)
        except:
            pass

    def _cache_invalidate(self):
        linkCache.invalidate()

    def batch_start(self):
        self.ipbatcbuf = ''
        self.ipbatch = True
        self.ipbatch_pause = False

    def add_to_batch(self, cmd):
        self.ipbatchbuf += cmd + '\n'

    def batch_pause(self):
        self.ipbatch_pause = True

    def batch_resume(self):
        self.ipbatch_pause = False

    def batch_commit(self):
        if not self.ipbatchbuf:
            return
        try:
            self.exec_command_talk_stdin('ip -force -batch -',
                    stdinbuf=self.ipbatchbuf)
        except Exception:
            raise
        finally:
            self.ipbatchbuf = ''
            self.ipbatch = False
            self.ipbatch_pause = False

    def addr_show(self, ifacename=None):
        if ifacename:
            if not self.link_exists(ifacename):
                return
            return self.exec_commandl(['ip','-o', 'addr', 'show', 'dev',
                    '%s' %ifacename])
        else:
            return self.exec_commandl(['ip', '-o', 'addr', 'show'])

    def link_show(self, ifacename=None):
        if ifacename:
            return self.exec_commandl(['ip', '-o', '-d', 'link',
                    'show', 'dev', '%s' %ifacename])
        else:
            return self.exec_commandl(['ip', '-o', '-d', 'link', 'show'])

    def addr_add(self, ifacename, address, broadcast=None,
                    peer=None, scope=None, preferred_lifetime=None):
        if not address:
            return
        cmd = 'addr add %s' %address
        if broadcast:
            cmd += ' broadcast %s' %broadcast
        if peer:
            cmd += ' peer %s' %peer
        if scope:
            cmd += ' scope %s' %scope
        if preferred_lifetime:
            cmd += ' preferred_lft %s' %preferred_lifetime
        cmd += ' dev %s' %ifacename
        if self.ipbatch and not self.ipbatch_pause:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip ' + cmd)
        self._cache_update([ifacename, 'addrs', address], {})

    def addr_del(self, ifacename, address, broadcast=None,
                    peer=None, scope=None):
        """ Delete ipv4 address """
        if not address:
            return
        if not self._cache_get('addr', [ifacename, 'addrs', address]):
            return
        cmd = 'addr del %s' %address
        if broadcast:
            cmd += 'broadcast %s' %broadcast
        if peer:
            cmd += 'peer %s' %peer
        if scope:
            cmd += 'scope %s' %scope
        cmd += ' dev %s' %ifacename
        self.exec_command('ip ' + cmd)
        self._cache_delete([ifacename, 'addrs', address])

    def addr_flush(self, ifacename):
        cmd = 'addr flush dev %s' %ifacename
        if self.ipbatch and not self.ipbatch_pause:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip ' + cmd)
        self._cache_delete([ifacename, 'addrs'])

    def del_addr_all(self, ifacename, skip_addrs=[]):
        if not skip_addrs: skip_addrs = []
        runningaddrsdict = self.addr_get(ifacename)
        try:
            # XXX: ignore errors. Fix this to delete secondary addresses
            # first
            [self.addr_del(ifacename, a) for a in
                set(runningaddrsdict.keys()).difference(skip_addrs)]
        except:
            # ignore errors
            pass

    def addr_get(self, ifacename, details=True):
        addrs = self._cache_get('addr', [ifacename, 'addrs'])
        if not addrs:
            return None
        if details:
            return addrs
        return addrs.keys()

    def addr_add_multiple(self, ifacename, addrs, purge_existing=False):
        # purges address
        if purge_existing:
            # if perfmode is not set and also if iface has no sibling
            # objects, purge addresses that are not present in the new
            # config
            runningaddrs = self.addr_get(ifacename, details=False)
            if addrs == runningaddrs:
                return
            try:
                # if primary address is not same, there is no need to keep any.
                # reset all addresses
                if (addrs and runningaddrs and
                        (addrs[0] != runningaddrs[0])):
                    self.del_addr_all(ifacename)
                else:
                    self.del_addr_all(ifacename, addrs)
            except Exception as e:
                self.log_warn(str(e))
        for a in addrs:
            try:
                self.addr_add(ifacename, a)
            except Exception as e:
                self.logger.error(str(e))

    def _link_set_ifflag(self, ifacename, value):
        # Dont look at the cache, the cache may have stale value
        # because link status can be changed by external
        # entity (One such entity is ifupdown main program)
        cmd = 'link set dev %s %s' %(ifacename, value.lower())
        if self.ipbatch:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip ' + cmd)

    def link_up(self, ifacename):
        self._link_set_ifflag(ifacename, 'UP')

    def link_down(self, ifacename):
        self._link_set_ifflag(ifacename, 'DOWN')

    def link_set(self, ifacename, key, value=None, force=False):
        if not force:
            if (key not in ['master', 'nomaster'] and
                self._cache_check('link', [ifacename, key], value)):
                return
        cmd = 'link set dev %s %s' %(ifacename, key)
        if value:
            cmd += ' %s' %value
        if self.ipbatch:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip ' + cmd)
        if key not in ['master', 'nomaster']:
            self._cache_update([ifacename, key], value)

    def link_set_hwaddress(self, ifacename, hwaddress, force=False):
        if not force:
            if self._cache_check('link', [ifacename, 'hwaddress'], hwaddress):
               return
        self.link_down(ifacename)
        cmd = 'link set dev %s address %s' %(ifacename, hwaddress)
        if self.ipbatch:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip ' + cmd)
        self.link_up(ifacename)
        self._cache_update([ifacename, 'hwaddress'], hwaddress)

    def link_set_alias(self, ifacename, alias):
        self.exec_commandl(['ip', 'link', 'set', 'dev',
                    ifacename, 'alias', alias])

    def link_get_alias(self, ifacename):
        return self.read_file_oneline('/sys/class/net/%s/ifalias'
                    %ifacename)

    def link_isloopback(self, ifacename):
        flags = self._cache_get('link', [ifacename, 'flags'])
        if not flags:
            return
        if 'LOOPBACK' in flags:
            return True
        return False

    def link_get_status(self, ifacename):
        return self._cache_get('link', [ifacename, 'ifflag'], refresh=True)

    def route_add_gateway(self, ifacename, gateway, metric=None):
        if not gateway:
           return
        cmd = 'ip route add default via %s' %gateway
        # Add metric
        if metric:
            cmd += 'metric %s' %metric
        cmd += ' dev %s' %ifacename
        self.exec_command(cmd)

    def route_del_gateway(self, ifacename, gateway, metric=None):
        # delete default gw
        if not gateway:
            return
        cmd = 'ip route del default via %s' %gateway
        if metric:
            cmd += ' metric %s' %metric
        cmd += ' dev %s' %ifacename
        self.exec_command(cmd)

    def route6_add_gateway(self, ifacename, gateway):
        if not gateway:
            return
        return self.exec_command('ip -6 route add default via %s' %gateway +
                                 ' dev %s' %ifacename)

    def route6_del_gateway(self, ifacename, gateway):
        if not gateway:
            return
        return self.exec_command('ip -6 route del default via %s' %gateway +
                                 'dev %s' %ifacename)

    def link_create_vlan(self, vlan_device_name, vlan_raw_device, vlanid):
        if self.link_exists(vlan_device_name):
            return
        self.exec_command('ip link add link %s' %vlan_raw_device +
                          ' name %s' %vlan_device_name +
                          ' type vlan id %d' %vlanid)
        self._cache_update([vlan_device_name], {})

    def link_create_vlan_from_name(self, vlan_device_name):
        v = vlan_device_name.split('.')
        if len(v) != 2:
            self.logger.warn('invalid vlan device name %s' %vlan_device_name)
            return 
        self.link_create_vlan(vlan_device_name, v[0], v[1])

    def link_create_macvlan(self, name, linkdev, mode='private'):
        if self.link_exists(name):
            return
        cmd = ('link add link %s' %linkdev +
                          ' name %s' %name +
                          ' type macvlan mode %s' %mode)
        if self.ipbatch and not self.ipbatch_pause:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip %s' %cmd)
        self._cache_update([name], {})

    def get_vxlan_peers(self, dev):
        cmd = 'bridge fdb show brport %s' % dev
        cur_peers = []
        try:
            ps = subprocess.Popen((cmd).split(), stdout=subprocess.PIPE, close_fds=True)
            output = subprocess.check_output(('grep', '00:00:00:00:00:00'), stdin=ps.stdout)
            ps.wait()
            try:
                ppat = re.compile('\s+dst\s+(\d+.\d+.\d+.\d+)\s+')
                for l in output.split('\n'):
                    m = ppat.search(l)
                    if m:
                        cur_peers.append(m.group(1))
            except:
                self.logger.warn('error parsing ip link output')
                pass
        except subprocess.CalledProcessError as e:
            if e.returncode != 1:
                self.logger.error(str(e))

        return cur_peers

    def link_create_vxlan(self, name, vxlanid,
                          localtunnelip=None,
                          svcnodeips=None,
                          remoteips=None,
                          learning='on',
                          ageing=None):
        if svcnodeips and remoteips:
            raise Exception("svcnodeip and remoteip is mutually exclusive")
        args = ''
        if localtunnelip:
            args += ' local %s' %localtunnelip
        if svcnodeips:
            for s in svcnodeips:
                args += ' svcnode %s' %s
        if ageing:
            args += ' ageing %s' %ageing
        if learning == 'off':
            args += ' nolearning'

        if self.link_exists(name):
            cmd = 'link set dev %s type vxlan dstport %d' %(name, VXLAN_UDP_PORT)
        else:
            cmd = 'link add dev %s type vxlan id %s dstport %d' %(name, vxlanid, VXLAN_UDP_PORT)
        cmd += args

        if self.ipbatch and not self.ipbatch_pause:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip %s' %cmd)

        # figure out the diff for remotes and do the bridge fdb updates
        cur_peers = set(self.get_vxlan_peers(name))
        if remoteips:
            new_peers = set(remoteips)
            del_list = cur_peers.difference(new_peers)
            add_list = new_peers.difference(cur_peers)
        else:
            del_list = cur_peers
            add_list = []

        try:
            for addr in del_list:
                self.bridge_fdb_del(name, '00:00:00:00:00:00', None, True, addr)
        except:
            pass

        try:
            for addr in add_list:
                self.bridge_fdb_append(name, '00:00:00:00:00:00', None, True, addr)
        except:
            pass

        # XXX: update linkinfo correctly
        self._cache_update([name], {})

    def link_exists(self, ifacename):
        if self.DRYRUN:
            return True
        return os.path.exists('/sys/class/net/%s' %ifacename)

    def is_vlan_device_by_name(self, ifacename):
        if re.search(r'\.', ifacename):
            return True
        return False

    def route_add(self, route):
        self.exec_command('ip route add ' + route)

    def route6_add(self, route):
        self.exec_command('ip -6 route add ' + route)

    def get_vlandev_attrs(self, ifacename):
        return (self._cache_get('link', [ifacename, 'linkinfo', 'link']),
                self._cache_get('link', [ifacename, 'linkinfo', 'vlanid']))

    def get_vxlandev_attrs(self, ifacename):
        return self._cache_get('link', [ifacename, 'linkinfo'])

    def link_get_mtu(self, ifacename):
        return self._cache_get('link', [ifacename, 'mtu'])

    def link_get_hwaddress(self, ifacename):
        address = self._cache_get('link', [ifacename, 'hwaddress'])
        # newly created logical interface addresses dont end up in the cache
        # read hwaddress from sysfs file for these interfaces
        if not address:
            address = self.read_file_oneline('/sys/class/net/%s/address'
                                             %ifacename)
        return address

    def link_create(self, ifacename, type, link=None):
        if self.link_exists(ifacename):
            return
        cmd = 'link add'
        if link:
            cmd += ' link %s' %link
        cmd += ' name %s type %s' %(ifacename, type)
        if self.ipbatch and not self.ipbatch_pause:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip %s' %cmd)
        self._cache_update([ifacename], {})

    def link_delete(self, ifacename):
        if not self.link_exists(ifacename):
            return
        cmd = 'link del %s' %ifacename
        if self.ipbatch and not self.ipbatch_pause:
            self.add_to_batch(cmd)
        else:
            self.exec_command('ip %s' %cmd)
        self._cache_invalidate()

    def bridge_port_vids_add(self, bridgeportname, vids):
        [self.exec_command('bridge vlan add vid %s dev %s'
                          %(v, bridgeportname)) for v in vids]

    def bridge_port_vids_del(self, bridgeportname, vids):
        if not vids:
            return
        [self.exec_command('bridge vlan del vid %s dev %s'
                          %(v, bridgeportname)) for v in vids]

    def bridge_port_vids_flush(self, bridgeportname):
        self.exec_command('bridge vlan del vid %s dev %s'
                          %(vid, bridgeportname))

    def bridge_port_vids_get(self, bridgeportname):
        self.exec_command('/sbin/bridge vlan show %s' %bridgeportname)
        bridgeout = self.exec_command('/sbin/bridge vlan show dev %s'
                                      %bridgeportname)
        if not bridgeout: return []
        brvlanlines = bridgeout.readlines()[2:]
        vids = [l.strip() for l in brvlanlines]
        return [vid for v in vids if vid]

    def bridge_port_vids_get_all(self):
        brvlaninfo = {}
        bridgeout = self.exec_command('/sbin/bridge vlan show')
        if not bridgeout: return brvlaninfo
        brvlanlines = bridgeout.splitlines()
        brportname=None
        for l in brvlanlines[1:]:
            if l and l[0] not in [' ', '\t']:
                brportname = None
            l=l.strip()
            if not l:
                brportname=None
                continue
            if 'PVID' in l:
                attrs = l.split()
                brportname = attrs[0]
                brvlaninfo[brportname] = {'pvid' : attrs[1],
                                          'vlan' : []}
            elif brportname:
                if 'Egress Untagged' not in l:
                    brvlaninfo[brportname]['vlan'].append(l)
            elif not brportname:
                attrs = l.split()
                if attrs[1] == 'None' or 'Egress Untagged' in attrs[1]:
                    continue
                brportname = attrs[0]
                brvlaninfo[brportname] = {'vlan' : [attrs[1]]}
        return brvlaninfo

    def bridge_port_pvid_add(self, bridgeportname, pvid):
        self.exec_command('bridge vlan add vid %s untagged pvid dev %s'
                          %(pvid, bridgeportname))

    def bridge_port_pvid_del(self, bridgeportname, pvid):
        self.exec_command('bridge vlan del vid %s untagged pvid dev %s'
                          %(pvid, bridgeportname))

    def bridge_port_pvids_get(self, bridgeportname):
        return self.read_file_oneline('/sys/class/net/%s/brport/pvid'
                                      %bridgeportname)

    def bridge_vids_add(self, bridgeportname, vids, bridge=True):
        target = 'self' if bridge else ''
        [self.exec_command('bridge vlan add vid %s dev %s %s'
                          %(v, bridgeportname, target)) for v in vids]

    def bridge_vids_del(self, bridgeportname, vids, bridge=True):
        target = 'self' if bridge else ''
        [self.exec_command('bridge vlan del vid %s dev %s %s'
                          %(v, bridgeportname, target)) for v in vids]

    def bridge_fdb_add(self, dev, address, vlan=None, bridge=True, remote=None):
        target = 'self' if bridge else ''
        vlan_str = ''
        if vlan:
            vlan_str = 'vlan %s ' % vlan

        dst_str = ''
        if remote:
            dst_str = 'dst %s ' % remote

        self.exec_command('bridge fdb replace %s dev %s %s %s %s'
                          %(address, dev, vlan_str, target, dst_str))

    def bridge_fdb_append(self, dev, address, vlan=None, bridge=True, remote=None):
        target = 'self' if bridge else ''
        vlan_str = ''
        if vlan:
            vlan_str = 'vlan %s ' % vlan
 
        dst_str = ''
        if remote:
            dst_str = 'dst %s ' % remote

        self.exec_command('bridge fdb append %s dev %s %s %s %s'
                          %(address, dev, vlan_str, target, dst_str))

    def bridge_fdb_del(self, dev, address, vlan=None, bridge=True, remote=None):
        target = 'self' if bridge else ''
        vlan_str = ''
        if vlan:
            vlan_str = 'vlan %s ' % vlan

        dst_str = ''
        if remote:
            dst_str = 'dst %s ' % remote
        self.exec_command('bridge fdb del %s dev %s %s %s %s'
                          %(address, dev, vlan_str, target, dst_str))

    def bridge_is_vlan_aware(self, bridgename):
        filename = '/sys/class/net/%s/bridge/vlan_filtering' %bridgename
        if os.path.exists(filename) and self.read_file_oneline(filename) == '1':
            return True
        return False

    def bridge_port_get_bridge_name(self, bridgeport):
        filename = '/sys/class/net/%s/brport/bridge' %bridgeport
        try:
            return os.path.basename(os.readlink(filename))
        except:
            return None

    def bridge_port_exists(self, bridge, bridgeportname):
        try:
            return os.path.exists('/sys/class/net/%s/brif/%s'
                                  %(bridge, bridgeportname))
        except Exception:
            return False

    def bridge_fdb_show_dev(self, dev):
        try:
            fdbs = {}
            output = self.exec_command('bridge fdb show dev %s' %dev)
            if output:
                for fdb_entry in output.splitlines():
                    try:
                        entries = fdb_entry.split()
                        fdbs.setdefault(entries[2], []).append(entries[0])
                    except:
                        self.logger.debug('%s: invalid fdb line \'%s\''
                                %(dev, fdb_entry))
                        pass
            return fdbs
        except Exception:
            return None

    def is_bridge(self, bridge):
        return os.path.exists('/sys/class/net/%s/bridge' %bridge)

    def is_link_up(self, ifacename):
        ret = False
        try:
            flags = self.read_file_oneline('/sys/class/net/%s/flags' %ifacename)
            iflags = int(flags, 16)
            if (iflags & 0x0001):
                ret = True
        except: 
            ret = False
            pass
        return ret

    def ip_route_get_dev(self, prefix):
        try:
            output = self.exec_command('ip route get %s' %prefix)
            if output:
               rline = output.splitlines()[0]
               if rline:
                    rattrs = rline.split()
                    return rattrs[rattrs.index('dev') + 1]
        except Exception as e:
            self.logger.debug('ip_route_get_dev: failed .. %s' %str(e))
            pass
        return None



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

from ifupdown.iface import *
from ifupdownaddons.modulebase import moduleBase
from ifupdownaddons.iproute2 import iproute2
import ifupdown.rtnetlink_api as rtnetlink_api
import logging
import re

class vlan(moduleBase):
    """  ifupdown2 addon module to configure vlans """

    _modinfo = {'mhelp' : 'vlan module configures vlan interfaces.' +
                        'This module understands vlan interfaces with dot ' +
                        'notations. eg swp1.100. Vlan interfaces with any ' +
                        'other names need to have raw device and vlan id ' +
                        'attributes',
                'attrs' : {
                        'vlan-raw-device' :
                            {'help' : 'vlan raw device'},
                        'vlan-id' :
                            {'help' : 'vlan id'}}}


    def __init__(self, *args, **kargs):
        moduleBase.__init__(self, *args, **kargs)
        self.ipcmd = None
        self._bridge_vids_query_cache = {}
        self._resv_vlan_range =  self._get_reserved_vlan_range()
        self.logger.debug('%s: using reserved vlan range %s'
                  %(self.__class__.__name__, str(self._resv_vlan_range)))

    def _is_vlan_device(self, ifaceobj):
        vlan_raw_device = ifaceobj.get_attr_value_first('vlan-raw-device')
        if vlan_raw_device:
            return True
        elif '.' in ifaceobj.name:
            return True
        return False

    def _get_vlan_id(self, ifaceobj):
        """ Derives vlanid from iface name
        
        Example:
            Returns 1 for ifname vlan0001 returns 1
            Returns 1 for ifname vlan1
            Returns 1 for ifname eth0.1

            Returns -1 if vlan id cannot be determined
        """
        vid_str = ifaceobj.get_attr_value_first('vlan-id')
        try:
            if vid_str: return int(vid_str)
        except:
            return -1

        if '.' in ifaceobj.name:
            vid_str = ifaceobj.name.split('.', 1)[1]
        elif ifaceobj.name.startswith('vlan'):
            vid_str = ifaceobj.name[4:]
        else:
            return -1
        try:
            vid = int(vid_str)
        except:
            return -1
        return vid

    def _is_vlan_by_name(self, ifacename):
        return '.' in ifacename

    def _get_vlan_raw_device_from_ifacename(self, ifacename):
        """ Returns vlan raw device from ifname
        Example:
            Returns eth0 for ifname eth0.100

            Returns None if vlan raw device name cannot
            be determined
        """
        vlist = ifacename.split('.', 1)
        if len(vlist) == 2:
            return vlist[0]
        return None

    def _get_vlan_raw_device(self, ifaceobj):
        vlan_raw_device = ifaceobj.get_attr_value_first('vlan-raw-device')
        if vlan_raw_device:
            return vlan_raw_device
        return self._get_vlan_raw_device_from_ifacename(ifaceobj.name)
        
    def get_dependent_ifacenames(self, ifaceobj, ifaceobjs_all=None):
        if not self._is_vlan_device(ifaceobj):
            return None
        ifaceobj.link_kind |= ifaceLinkKind.VLAN
        return [self._get_vlan_raw_device(ifaceobj)]

    def _bridge_vid_add_del(self, ifaceobj, bridgename, vlanid,
                            add=True):
        """ If the lower device is a vlan aware bridge, add/del the vlanid
        to the bridge """
        if self.ipcmd.bridge_is_vlan_aware(bridgename):
           if add:
              rtnetlink_api.rtnl_api.bridge_vlan(add=True, dev=bridgename,
                                                 vid=vlanid, master=False)
           else:
              rtnetlink_api.rtnl_api.bridge_vlan(add=False, dev=bridgename,
                                                 vid=vlanid, master=False)

    def _bridge_vid_check(self, ifaceobj, ifaceobjcurr, bridgename, vlanid):
        """ If the lower device is a vlan aware bridge, check if the vlanid
        is configured on the bridge """
        if not self.ipcmd.bridge_is_vlan_aware(bridgename):
            return
        vids = self._bridge_vids_query_cache.get(bridgename)
        if vids == None:
           vids = self.ipcmd.bridge_port_vids_get(bridgename)
           self._bridge_vids_query_cache[bridgename] = vids
        if not vids or vlanid not in vids:
            ifaceobjcurr.status = ifaceStatus.ERROR
            ifaceobjcurr.status_str = 'bridge vid error'

    def _up(self, ifaceobj):
        vlanid = self._get_vlan_id(ifaceobj)
        if vlanid == -1:
            raise Exception('could not determine vlanid')
        if self._handle_reserved_vlan(vlanid, ifaceobj.name):
           return
        vlanrawdevice = self._get_vlan_raw_device(ifaceobj)
        if not vlanrawdevice:
            raise Exception('could not determine vlan raw device')
        if not self.PERFMODE:
            if not self.ipcmd.link_exists(vlanrawdevice):
                raise Exception('rawdevice %s not present' %vlanrawdevice)
            if self.ipcmd.link_exists(ifaceobj.name):
                self._bridge_vid_add_del(ifaceobj, vlanrawdevice, vlanid)
                return
        rtnetlink_api.rtnl_api.create_vlan(vlanrawdevice,
                    ifaceobj.name, vlanid)
        self._bridge_vid_add_del(ifaceobj, vlanrawdevice, vlanid)
        if ifaceobj.addr_method == 'manual':
           rtnetlink_api.rtnl_api.link_set(ifaceobj.name, "up")

    def _down(self, ifaceobj):
        vlanid = self._get_vlan_id(ifaceobj)
        if vlanid == -1:
            raise Exception('could not determine vlanid')
        vlanrawdevice = self._get_vlan_raw_device(ifaceobj)
        if not vlanrawdevice:
            raise Exception('could not determine vlan raw device')
        if not self.PERFMODE and not self.ipcmd.link_exists(ifaceobj.name):
           return
        try:
            self.ipcmd.link_delete(ifaceobj.name)
            self._bridge_vid_add_del(ifaceobj, vlanrawdevice, vlanid, add=False)
        except Exception as e:
            self.log_warn(str(e))

    def _query_check(self, ifaceobj, ifaceobjcurr):
        if not self.ipcmd.link_exists(ifaceobj.name):
           return
        if not '.' in ifaceobj.name:
            # if vlan name is not in the dot format, check its running state
            (vlanrawdev, vlanid) = self.ipcmd.get_vlandev_attrs(ifaceobj.name)
            if vlanrawdev != ifaceobj.get_attr_value_first('vlan-raw-device'):
                ifaceobjcurr.update_config_with_status('vlan-raw-device',
                        vlanrawdev, 1)
            else:
                ifaceobjcurr.update_config_with_status('vlan-raw-device',
                        vlanrawdev, 0)
            if vlanid != ifaceobj.get_attr_value_first('vlan-id'):
                ifaceobjcurr.update_config_with_status('vlan-id', vlanid, 1)
            else:
                ifaceobjcurr.update_config_with_status('vlan-id',
                        vlanid, 0)
            self._bridge_vid_check(ifaceobj, ifaceobjcurr, vlanrawdev, vlanid)

    def _query_running(self, ifaceobjrunning):
        if not self.ipcmd.link_exists(ifaceobjrunning.name):
            return
        if not self.ipcmd.get_vlandev_attrs(ifaceobjrunning.name):
            return
        # If vlan name is not in the dot format, get the
        # vlan dev and vlan id
        if not '.' in ifaceobjrunning.name:
            (vlanrawdev, vlanid) = self.ipcmd.get_vlandev_attrs(ifaceobjrunning.name)
            ifaceobjrunning.update_config_dict({(k, v) for k, v in
                                                {'vlan-raw-device' : vlanrawdev,
                                                 'vlan-id' : vlanid}.items()
                                                if v})

    _run_ops = {'pre-up' : _up,
               'post-down' : _down,
               'query-checkcurr' : _query_check,
               'query-running' : _query_running}

    def get_ops(self):
        """ returns list of ops supported by this module """
        return self._run_ops.keys()

    def _init_command_handlers(self):
        if not self.ipcmd:
            self.ipcmd = iproute2(**self.get_flags())
        

    def run(self, ifaceobj, operation, query_ifaceobj=None, **extra_args):
        """ run vlan configuration on the interface object passed as argument

        Args:
            **ifaceobj** (object): iface object

            **operation** (str): any of 'pre-up', 'post-down', 'query-checkcurr',
                                 'query-running'
        Kwargs:
            **query_ifaceobj** (object): query check ifaceobject. This is only
                valid when op is 'query-checkcurr'. It is an object same as
                ifaceobj, but contains running attribute values and its config
                status. The modules can use it to return queried running state
                of interfaces. status is success if the running state is same
                as user required state in ifaceobj. error otherwise.
        """
        if ifaceobj.type == ifaceType.BRIDGE_VLAN:
            return
        op_handler = self._run_ops.get(operation)
        if not op_handler:
            return
        if (operation != 'query-running' and
                not self._is_vlan_device(ifaceobj)):
            return
        self._init_command_handlers()
        if operation == 'query-checkcurr':
            op_handler(self, ifaceobj, query_ifaceobj)
        else:
            op_handler(self, ifaceobj)



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

import logging
import subprocess
import re
import io
from ifupdown.iface import *
from cache import *

#import timeit
import time
import logging

def profile(func):
    def wrap(*args, **kwargs):
        started_at = time.time()
        result = func(*args, **kwargs)
        print (str(func))
        print (time.time() - started_at)
        return result
    return wrap

class utilsBase(object):
    """ Base class for ifupdown addon utilities """

    def __init__(self, *args, **kargs):
        modulename = self.__class__.__name__
        self.logger = logging.getLogger('ifupdown.' + modulename)
        self.FORCE = kargs.get('force', False)
        self.DRYRUN = kargs.get('dryrun', False)
        self.NOWAIT = kargs.get('nowait', False)
        self.PERFMODE = kargs.get('perfmode', False)
        self.CACHE = kargs.get('cache', False)

    def exec_commandl(self, cmdl, cmdenv=None):
        """ Executes command """

        cmd_returncode = 0
        cmdout = ''
        try:
            self.logger.info('executing ' + ' '.join(cmdl))
            if self.DRYRUN:
                return cmdout
            ch = subprocess.Popen(cmdl,
                    stdout=subprocess.PIPE,
                    shell=False, env=cmdenv,
                    stderr=subprocess.STDOUT,
                    close_fds=True)
            cmdout = ch.communicate()[0]
            cmd_returncode = ch.wait()
        except OSError as e:
            raise Exception('failed to execute cmd \'%s\' (%s)'
                            %(' '.join(cmdl), str(e)))
        if cmd_returncode != 0:
            raise Exception('failed to execute cmd \'%s\''
                 %' '.join(cmdl) + '(' + cmdout.strip('\n ') + ')')
        return cmdout

    def exec_command(self, cmd, cmdenv=None):
        """ Executes command given as string in the argument cmd """

        return self.exec_commandl(cmd.split(), cmdenv)

    def exec_command_talk_stdin(self, cmd, stdinbuf):
        """ Executes command and writes to stdin of the process """
        cmd_returncode = 0
        cmdout = ''
        try:
            self.logger.info('executing %s [%s]' %(cmd, stdinbuf))
            if self.DRYRUN:
                return cmdout
            ch = subprocess.Popen(cmd.split(),
                    stdout=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                    shell=False,
                    stderr=subprocess.STDOUT,
                    close_fds=True)
            cmdout = ch.communicate(input=stdinbuf)[0]
            cmd_returncode = ch.wait()
        except OSError as e:
            raise Exception('failed to execute cmd \'%s\' (%s)'
                            %(cmd, str(e)))
        if cmd_returncode != 0:
            raise Exception('failed to execute cmd \'%s [%s]\''
                %(cmd, stdinbuf) + '(' + cmdout.strip('\n ') + ')')
        return cmdout

    def subprocess_check_output(self, cmdl):
        self.logger.info('executing ' + ' '.join(cmdl))
        if self.DRYRUN:
            return
        try:
            return subprocess.check_output(cmdl, stderr=subprocess.STDOUT)
        except Exception as e:
            raise Exception('failed to execute cmd \'%s\' (%s)'
                        %(' '.join(cmdl), e.output))

    def subprocess_check_call(self, cmdl):
        """ subprocess check_call implementation using popen
        
        Uses popen because it needs the close_fds argument
        """

        cmd_returncode = 0
        try:
            self.logger.info('executing ' + ' '.join(cmdl))
            if self.DRYRUN:
                return
            ch = subprocess.Popen(cmdl,
                    stdout=None,
                    shell=False,
                    stderr=None,
                    close_fds=True)
            cmd_returncode = ch.wait()
        except Exception as e:
            raise Exception('failed to execute cmd \'%s\' (%s)'
                            %(' '.join(cmdl), str(e)))
        if cmd_returncode != 0:
            raise Exception('failed to execute cmd \'%s\''
                 %' '.join(cmdl))
        return

    def write_file(self, filename, strexpr):
        try:
            self.logger.info('writing \'%s\'' %strexpr +
                ' to file %s' %filename)
            if self.DRYRUN:
                return 0
            with open(filename, 'w') as f:
                f.write(strexpr)
        except IOError as e:
            self.logger.warn('error writing to file %s'
                %filename + '(' + str(e) + ')')
            return -1
        return 0

    def read_file(self, filename):
        try:
            self.logger.debug('reading \'%s\'' %filename)
            with open(filename, 'r') as f:
                return f.readlines()
        except:
            return None
        return None

    def read_file_oneline(self, filename):
        try:
            self.logger.debug('reading \'%s\'' %filename)
            with open(filename, 'r') as f:
                return f.readline().strip('\n')
        except:
            return None
        return None

    def sysctl_set(self, variable, value):
        self.exec_command('sysctl %s=' %variable + '%s' %value)

    def sysctl_get(self, variable):
        return self.exec_command('sysctl %s' %variable).split('=')[1].strip()



#!/usr/bin/python

import argparse
import sys
import subprocess

""" This script prints to stdout /etc/network/interfaces entries for
    requested interfaces.

    Currently it supports generation of interfaces(5) section for all
    swp interfaces on the system. And also an interface section 
    for a bridge with all swp ports.

    Example use of this script:

    generate the swp_defaults file:
    (bkup existing /etc/network/interfaces.d/swp_defaults file if one exists)

    #generate_interfaces.py -s > /etc/network/interfaces.d/swp_defaults

    User -m option if you want the new swp_defaults to be auto merged
    with the contents from the old file, use -m option

    #generate_interfaces.py -s -m /etc/network/interfaces.d/swp_defaults > /etc/network/interfaces.d/swp_defaults.new

    Include the swp_defaults file in /etc/network/interfaces file
    (if not already there) using the source command as shown below:

    source /etc/network/interfaces.d/swp_defaults

"""

def get_swp_interfaces():
    porttab_path = '/var/lib/cumulus/porttab'
    ports = []

    ptfile = open(porttab_path, 'r')
    for line in ptfile.readlines():
        line = line.strip()
        if '#' in line:
            continue
        try:
            ports.append(line.split()[0])
        except ValueError:
            continue
    return ports

def print_swp_defaults_header():
    print '''
# ** This file is autogenerated by /usr/share/doc/python-ifupdown2/generate_interfaces.py **
#
# This is /etc/network/interfaces section for all available swp
# ports on the system.
#
# To include this file in the main /etc/network/interfaces file,
# copy this file under /etc/network/interfaces.d/ and use the
# source line in the /etc/network/interfaces file.
#
# example entry in /etc/network/interfaces:
#   source /etc/network/interfaces.d/<filename>
#
# See manpage interfaces(5) for details.
'''

def print_bridge_untagged_defaults_header():
    print '''
# ** This file is autogenerated by /usr/share/doc/python-ifupdown2/generate_interfaces.py **
#
# This is /etc/network/interfaces section for a bridge device with all swp
# ports in the system.
#
# To include this file in the main /etc/network/interfaces file,
# copy this file under /etc/network/interfaces.d/ and use the
# source line in the /etc/network/interfaces file as shown below.
# details.
#
# example entry in /etc/network/interfaces:
#   source /etc/network/interfaces.d/filename
#
# See manpage interfaces(5) for details
'''

def interfaces_print_swp_default(swp_intf):
    outbuf = None
    if args.mergefile:
        try:
            cmd = ['/sbin/ifquery', '%s' %swp_intf, '-i', '%s' %args.mergefile]
            outbuf = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except Exception, e:
            # no interface found gen latest
            pass
    if not outbuf:
        outbuf = 'auto %s\niface %s\n\n' %(swp_intf, swp_intf)
    return outbuf

def interfaces_print_swp_defaults(swp_intfs):
    print_swp_defaults_header()
    outbuf = ''
    for i in swp_intfs:
        outbuf += interfaces_print_swp_default(i)
    print outbuf

def interfaces_print_bridge_default(swp_intfs):
    print_bridge_untagged_defaults_header()
    outbuf = 'auto bridge-untagged\n' 
    outbuf += 'iface bridge-untagged\n'
    outbuf += '  bridge-ports \\\n'
    linen = 5
    ports = ''
    for i in range(0, len(swp_intfs), linen):
        if ports:
            ports += ' \\\n'
        ports += '      %s' %(' '.join(swp_intfs[i:i+linen]))
    outbuf += ports
    print outbuf

def populate_argparser(argparser):
    group = argparser.add_mutually_exclusive_group(required=False)
    group.add_argument('-s', '--swp-defaults', action='store_true',
                       dest='swpdefaults', help='generate swp defaults file')
    group.add_argument('-b', '--bridge-default', action='store_true',
                       dest='bridgedefault',
                       help='generate default untagged bridge')
    argparser.add_argument('-m', '--merge', dest='mergefile', help='merge ' +
                           'new generated iface content with the old one')

argparser =  argparse.ArgumentParser(description='ifupdown interfaces file gen helper')
populate_argparser(argparser)
args = argparser.parse_args(sys.argv[1:])

if not args.swpdefaults and not args.bridgedefault:
    argparser.print_help()
    exit(1)

if args.bridgedefault and args.mergefile:
    print 'error: mergefile option currently only supported with -s'
    argparser.print_help()
    exit(1)

swp_intfs = get_swp_interfaces()
if not swp_intfs:
    print 'error: no ports found'
    exit(1)

if args.swpdefaults:
    interfaces_print_swp_defaults(swp_intfs)
elif args.bridgedefault:
    interfaces_print_bridge_default(swp_intfs)
else:
    argparser.print_help()



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#
# template --
#    helper class to render templates
#

import logging
import traceback
from utils import *

class templateEngine():
    """ provides template rendering methods """

    def __init__(self, template_engine, template_enable='0',
                 template_lookuppath=None):
        self.logger = logging.getLogger('ifupdown.' +
                    self.__class__.__name__)
        self.tclass = None
        self.tclassargs = {}
        self.render = self._render_default
        if template_enable == '0':
            return
        if template_engine == 'mako':
            try:
                self.tclass = utils.importName('mako.template', 'Template')
            except Exception as e:
                self.logger.warn('unable to load template engine %s (%s)'
                        %(template_engine, str(e)))
                pass
            if template_lookuppath:
                try:
                    self.logger.debug('setting template lookuppath to %s'
                            %template_lookuppath)
                    lc = utils.importName('mako.lookup', 'TemplateLookup')
                    self.tclassargs['lookup'] = lc(
                                directories=template_lookuppath.split(':'))
                except Exception as e:
                    self.logger.warn('unable to set template lookup path' +
                                ' %s (%s)' %(template_lookuppath, str(e)))
                    pass
            self.render = self._render_mako
        else:
            self.logger.info('skip template processing.., ' +
                    'template engine not found')

    def _render_default(self, textdata):
        return textdata

    def _render_mako(self, textdata):
        """ render textdata passed as argument using mako

        Returns rendered textdata """

        if not self.tclass:
            return textdata
        self.logger.info('template processing on interfaces file ...')
        t = self.tclass(text=textdata, output_encoding='utf-8',
                     lookup=self.tclassargs.get('lookup'))
        return t.render()



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

from ifupdown.iface import *
from utilsbase import *
import os
import re
import logging
from cache import *

class brctl(utilsBase):
    """ This class contains helper functions to interact with the bridgeutils
    commands """

    _cache_fill_done = False

    def __init__(self, *args, **kargs):
        utilsBase.__init__(self, *args, **kargs)
        if self.CACHE and not brctl._cache_fill_done:
            if os.path.exists('/sbin/brctl'):
                self._bridge_fill()
            self._cache_fill_done = True

    def _bridge_get_mcattrs_from_sysfs(self, bridgename):
        mcattrs = {}
        mcattrmap = {'mclmc': 'multicast_last_member_count',
                     'mcrouter': 'multicast_router',
                     'mcsnoop' : 'multicast_snooping',
                     'mcsqc' : 'multicast_startup_query_count',
                     'mcqifaddr' : 'multicast_query_use_ifaddr',
                     'mcquerier' : 'multicast_querier',
                     'hashel' : 'hash_elasticity',
                     'hashmax' : 'hash_max',
                     'mclmi' : 'multicast_last_member_interval',
                     'mcmi' : 'multicast_membership_interval',
                     'mcqpi' : 'multicast_querier_interval',
                     'mcqi' : 'multicast_query_interval',
                     'mcqri' : 'multicast_query_response_interval',
                     'mcsqi' : 'multicast_startup_query_interval'}

        mcattrsdivby100 = ['mclmi', 'mcmi', 'mcqpi', 'mcqi', 'mcqri', 'mcsqi']

        for m, s in mcattrmap.items():
            n = self.read_file_oneline('/sys/class/net/%s/bridge/%s'
                                    %(bridgename, s))
            if m in mcattrsdivby100:
                try:
                    v = int(n) / 100
                    mcattrs[m] = str(v)
                except Exception as e:
                    self.logger.warn('error getting mc attr %s (%s)'
                                     %(m, str(e)))
                    pass
            else:
                mcattrs[m] = n
        return mcattrs

    def _bridge_attrs_fill(self, bridgename):
        battrs = {}
        bports = {}

        brout = self.exec_command('/sbin/brctl showstp %s' %bridgename)
        chunks = re.split(r'\n\n', brout, maxsplit=0, flags=re.MULTILINE)

        try:
            # Get all bridge attributes
            broutlines = chunks[0].splitlines()
            #battrs['pathcost'] = broutlines[3].split('path cost')[1].strip()
            battrs['maxage'] = broutlines[4].split(
                                'bridge max age')[1].strip().replace('.00', '')
            battrs['hello'] = broutlines[5].split(
                                'bridge hello time')[1].strip().replace('.00',
                                                                        '')
            battrs['fd'] = broutlines[6].split(
                                    'bridge forward delay')[1].strip(
                                            ).replace('.00', '')
            battrs.update(self._bridge_get_mcattrs_from_sysfs(bridgename))

            # XXX: comment this out until mc attributes become available
            # with brctl again
            #battrs['hashel'] = broutlines[10].split('hash elasticity')[1].split()[0].strip()
            #battrs['hashmax'] = broutlines[10].split('hash max')[1].strip()
            #battrs['mclmc'] = broutlines[11].split('mc last member count')[1].split()[0].strip()
            #battrs['mciqc'] = broutlines[11].split('mc init query count')[1].strip()
            #battrs['mcrouter'] = broutlines[12].split('mc router')[1].split()[0].strip()
            ##battrs['mcsnoop'] = broutlines[12].split('mc snooping')[1].strip()
            #battrs['mclmt'] = broutlines[13].split('mc last member timer')[1].split()[0].strip()
        except Exception as e:
            self.logger.warn(str(e))
            pass

        linkCache.update_attrdict([bridgename, 'linkinfo'], battrs)

        for cidx in range(1, len(chunks)):
            bpout = chunks[cidx].lstrip('\n')
            if not bpout or bpout[0] == ' ':
                continue
            bplines = bpout.splitlines()
            pname = bplines[0].split()[0]
            bportattrs = {}
            try:
                bportattrs['pathcost'] = bplines[2].split(
                                            'path cost')[1].strip()
                bportattrs['fdelay'] = bplines[4].split(
                                            'forward delay timer')[1].strip()
                bportattrs['mcrouter'] = self.read_file_oneline(
                         '/sys/class/net/%s/brport/multicast_router' %pname)
                bportattrs['mcfl'] = self.read_file_oneline(
                         '/sys/class/net/%s/brport/multicast_fast_leave' %pname)

                #bportattrs['mcrouters'] = bplines[6].split('mc router')[1].split()[0].strip()
                #bportattrs['mc fast leave'] = bplines[6].split('mc fast leave')[1].strip()
            except Exception as e:
                self.logger.warn(str(e))
                pass
            bports[pname] = bportattrs
            linkCache.update_attrdict([bridgename, 'linkinfo', 'ports'], bports)

    def _bridge_fill(self, bridgename=None, refresh=False):
        try:
            # if cache is already filled, return
            linkCache.get_attr([bridgename, 'linkinfo', 'fd'])
            return
        except:
            pass
        if not bridgename:
            brctlout = self.exec_command('/sbin/brctl show')
        else:
            brctlout = self.exec_command('/sbin/brctl show ' + bridgename)
        if not brctlout:
            return

        for bline in brctlout.splitlines()[1:]:
            bitems = bline.split()
            if len(bitems) < 2:
                continue
            try:
                linkCache.update_attrdict([bitems[0], 'linkinfo'],
                                      {'stp' : bitems[2]})
            except KeyError:
                linkCache.update_attrdict([bitems[0]], 
                                {'linkinfo' : {'stp' : bitems[2]}})
            self._bridge_attrs_fill(bitems[0])

    def _cache_get(self, attrlist, refresh=False):
        try:
            if self.DRYRUN:
                return None
            if self.CACHE:
                if not self._cache_fill_done: 
                    self._bridge_fill()
                    self._cache_fill_done = True
                    return linkCache.get_attr(attrlist)
                if not refresh:
                    return linkCache.get_attr(attrlist)
            self._bridge_fill(attrlist[0], refresh)
            return linkCache.get_attr(attrlist)
        except Exception as e:
            self.logger.debug('_cache_get(%s) : [%s]'
                    %(str(attrlist), str(e)))
            pass
        return None

    def _cache_check(self, attrlist, value, refresh=False):
        try:
            attrvalue = self._cache_get(attrlist, refresh)
            if attrvalue and attrvalue == value:
                return True
        except Exception as e:
            self.logger.debug('_cache_check(%s) : [%s]'
                    %(str(attrlist), str(e)))
            pass
        return False

    def _cache_update(self, attrlist, value):
        if self.DRYRUN: return
        try:
            linkCache.add_attr(attrlist, value)
        except:
            pass

    def _cache_delete(self, attrlist):
        if self.DRYRUN: return
        try:
            linkCache.del_attr(attrlist)
        except:
            pass

    def _cache_invalidate(self):
        if self.DRYRUN: return
        linkCache.invalidate()

    def create_bridge(self, bridgename):
        if self.bridge_exists(bridgename):
            return
        self.exec_command('/sbin/brctl addbr %s' %bridgename)
        self._cache_update([bridgename], {})

    def delete_bridge(self, bridgename):
        if not self.bridge_exists(bridgename):
            return
        self.exec_command('/sbin/brctl delbr %s' %bridgename)
        self._cache_invalidate()

    def add_bridge_port(self, bridgename, bridgeportname):
        """ Add port to bridge """
        ports = self._cache_get([bridgename, 'linkinfo', 'ports'])
        if ports and ports.get(bridgeportname):
            return
        self.exec_command('/sbin/brctl addif ' + bridgename + ' ' +
                          bridgeportname)
        self._cache_update([bridgename, 'linkinfo', 'ports',
                            bridgeportname], {})

    def delete_bridge_port(self, bridgename, bridgeportname):
        """ Delete port from bridge """
        ports = self._cache_get([bridgename, 'linkinfo', 'ports'])
        if not ports or not ports.get(bridgeportname):
            return
        self.exec_command('/sbin/brctl delif ' + bridgename + ' ' +
                          bridgeportname)
        self._cache_delete([bridgename, 'linkinfo', 'ports',
                           'bridgeportname'])

    def set_bridgeport_attrs(self, bridgename, bridgeportname, attrdict):
        portattrs = self._cache_get([bridgename, 'linkinfo',
                                       'ports', bridgeportname])
        if portattrs == None: portattrs = {}
        for k, v in attrdict.iteritems():
            if self.CACHE:
                curval = portattrs.get(k)
                if curval and curval == v:
                    continue
            self.exec_command('/sbin/brctl set%s %s %s %s'
                              %(k, bridgename, bridgeportname, v))

    def set_bridgeport_attr(self, bridgename, bridgeportname,
                            attrname, attrval):
        if self._cache_check([bridgename, 'linkinfo', 'ports',
                        bridgeportname, attrname], attrval):
            return
        self.exec_command('/sbin/brctl set%s %s %s %s' %(attrname, bridgename,
                          bridgeportname, attrval))

    def set_bridge_attrs(self, bridgename, attrdict):
        for k, v in attrdict.iteritems():
            if not v:
                continue
            if self._cache_check([bridgename, 'linkinfo', k], v):
                continue
            try:
                self.exec_command('/sbin/brctl set%s %s %s'
                                  %(k, bridgename, v))
            except Exception as e:
                self.logger.warn('%s: %s' %(bridgename, str(e)))
                pass

    def set_bridge_attr(self, bridgename, attrname, attrval):
        if self._cache_check([bridgename, 'linkinfo', attrname], attrval):
            return
        self.exec_command('/sbin/brctl set%s %s %s'
                          %(attrname, bridgename, attrval))

    def get_bridge_attrs(self, bridgename):
        return self._cache_get([bridgename, 'linkinfo'])

    def get_bridgeport_attrs(self, bridgename, bridgeportname):
        return self._cache_get([bridgename, 'linkinfo', 'ports',
                                      bridgeportname])

    def get_bridgeport_attr(self, bridgename, bridgeportname, attrname):
        return self._cache_get([bridgename, 'linkinfo', 'ports',
                                      bridgeportname, attrname])

    def set_stp(self, bridge, stp_state):
        self.exec_command('/sbin/brctl stp ' + bridge + ' ' + stp_state)

    def get_stp(self, bridge):
        sysfs_stpstate = '/sys/class/net/%s/bridge/stp_state' %bridge
        if not os.path.exists(sysfs_stpstate):
            return 'error'
        stpstate = self.read_file_oneline(sysfs_stpstate)
        if not stpstate:
            return 'error'
        try:
            if int(stpstate) > 0:
                return 'yes'
            elif int(stpstate) == 0:
                return 'no'
        except:
            return 'unknown'

    def conv_value_to_user(self, str):
        try:
            ret = int(str) / 100
        except:
            return None
        finally:
            return '%d' %ret

    def read_value_from_sysfs(self, filename, preprocess_func):
        value = self.read_file_oneline(filename)
        if not value:
            return None
        return preprocess_func(value)

    def set_ageing(self, bridge, ageing):
        self.exec_command('/sbin/brctl setageing ' + bridge + ' ' + ageing)

    def get_ageing(self, bridge):
        return self.read_value_from_sysfs('/sys/class/net/%s/bridge/ageing_time'
                                     %bridge, self.conv_value_to_user)

    def set_bridgeprio(self, bridge, bridgeprio):
        self.exec_command('/sbin/brctl setbridgeprio ' + bridge + ' ' +
                            bridgeprio)

    def get_bridgeprio(self, bridge):
        return self.read_file_oneline(
                       '/sys/class/net/%s/bridge/priority' %bridge)

    def set_fd(self, bridge, fd):
        self.exec_command('/sbin/brctl setfd ' + bridge + ' ' + fd)

    def get_fd(self, bridge):
        return self.read_value_from_sysfs(
                            '/sys/class/net/%s/bridge/forward_delay'
                            %bridge, self.conv_value_to_user)

    def set_gcint(self, bridge, gcint):
        #cmd = '/sbin/brctl setgcint ' + bridge + ' ' + gcint
        raise Exception('set_gcint not implemented')

    def set_hello(self, bridge, hello):
        self.exec_command('/sbin/brctl sethello ' + bridge + ' ' + hello)

    def get_hello(self, bridge):
        return self.read_value_from_sysfs('/sys/class/net/%s/bridge/hello_time'
                                          %bridge, self.conv_value_to_user)

    def set_maxage(self, bridge, maxage):
        self.exec_command('/sbin/brctl setmaxage ' + bridge + ' ' + maxage)

    def get_maxage(self, bridge):
        return self.read_value_from_sysfs('/sys/class/net/%s/bridge/max_age'
                                          %bridge, self.conv_value_to_user)

    def set_pathcost(self, bridge, port, pathcost):
        self.exec_command('/sbin/brctl setpathcost %s' %bridge + ' %s' %port +
                            ' %s' %pathcost)

    def get_pathcost(self, bridge, port):
        return self.read_file_oneline('/sys/class/net/%s/brport/path_cost'
                                        %port)

    def set_portprio(self, bridge, port, prio):
        self.exec_command('/sbin/brctl setportprio %s' %bridge + ' %s' %port +
                          ' %s' %prio)

    def get_portprio(self, bridge, port):
        return self.read_file_oneline('/sys/class/net/%s/brport/priority'
                                        %port)

    def set_hashmax(self, bridge, hashmax):
        self.exec_command('/sbin/brctl sethashmax %s' %bridge + ' %s' %hashmax)

    def get_hashmax(self, bridge):
        return self.read_file_oneline('/sys/class/net/%s/bridge/hash_max'
                                        %bridge)

    def set_hashel(self, bridge, hashel):
        self.exec_command('/sbin/brctl sethashel %s' %bridge + ' %s' %hashel)

    def get_hashel(self, bridge):
        return self.read_file_oneline('/sys/class/net/%s/bridge/hash_elasticity'
                                        %bridge)

    def set_mclmc(self, bridge, mclmc):
        self.exec_command('/sbin/brctl setmclmc %s' %bridge + ' %s' %mclmc)

    def get_mclmc(self, bridge):
        return self.read_file_oneline(
                    '/sys/class/net/%s/bridge/multicast_last_member_count'
                    %bridge)

    def set_mcrouter(self, bridge, mcrouter):
        self.exec_command('/sbin/brctl setmcrouter %s' %bridge +
                          ' %s' %mcrouter)

    def get_mcrouter(self, bridge):
        return self.read_file_oneline(
                    '/sys/class/net/%s/bridge/multicast_router' %bridge)

    def set_mcsnoop(self, bridge, mcsnoop):
        self.exec_command('/sbin/brctl setmcsnoop %s' %bridge +
                          ' %s' %mcsnoop)

    def get_mcsnoop(self, bridge):
        return self.read_file_oneline(
                    '/sys/class/net/%s/bridge/multicast_snooping' %bridge)

    def set_mcsqc(self, bridge, mcsqc):
        self.exec_command('/sbin/brctl setmcsqc %s' %bridge +
                          ' %s' %mcsqc)

    def get_mcsqc(self, bridge):
        return self.read_file_oneline(
                    '/sys/class/net/%s/bridge/multicast_startup_query_count'
                    %bridge)

    def set_mcqifaddr(self, bridge, mcqifaddr):
        self.exec_command('/sbin/brctl setmcqifaddr %s' %bridge +
                          ' %s' %mcqifaddr)

    def get_mcqifaddr(self, bridge):
        return self.read_file_oneline(
                 '/sys/class/net/%s/bridge/multicast_startup_query_use_ifaddr'
                 %bridge)

    def set_mcquerier(self, bridge, mcquerier):
        self.exec_command('/sbin/brctl setmcquerier %s' %bridge +
                          ' %s' %mcquerier)

    def get_mcquerier(self, bridge):
        return self.read_file_oneline(
                 '/sys/class/net/%s/bridge/multicast_querier' %bridge)

    def set_mcqv4src(self, bridge, vlan, mcquerier):
        if vlan == 0 or vlan > 4095:
            self.logger.warn('mcqv4src vlan \'%d\' invalid range' %vlan)
            return

        ip = mcquerier.split('.')
        if len(ip) != 4:
            self.logger.warn('mcqv4src \'%s\' invalid IPv4 address' %mcquerier)
            return
        for k in ip:
            if not k.isdigit() or int(k, 10) < 0 or int(k, 10) > 255:
                self.logger.warn('mcqv4src \'%s\' invalid IPv4 address' %mcquerier)
                return

        self.exec_command('/sbin/brctl setmcqv4src %s' %bridge +
                          ' %d %s' %(vlan, mcquerier)) 

    def del_mcqv4src(self, bridge, vlan):
        self.exec_command('/sbin/brctl delmcqv4src %s %d' %(bridge, vlan))

    def get_mcqv4src(self, bridge, vlan=None):
        mcqv4src = {}
        mcqout = self.exec_command('/sbin/brctl showmcqv4src %s' %bridge)
        if not mcqout: return None
        mcqlines = mcqout.splitlines()
        for l in mcqlines[1:]:
            l=l.strip()
            k, d, v = l.split('\t')
            if not k or not v:
                continue
            mcqv4src[k] = v
        if vlan:
            return mcqv4src.get(vlan)
        return mcqv4src

    def set_mclmi(self, bridge, mclmi):
        self.exec_command('/sbin/brctl setmclmi %s' %bridge +
                          ' %s' %mclmi)

    def get_mclmi(self, bridge):
        return self.read_file_oneline(
                 '/sys/class/net/%s/bridge/multicast_last_member_interval'
                 %bridge)

    def set_mcmi(self, bridge, mcmi):
        self.exec_command('/sbin/brctl setmcmi %s' %bridge +
                          ' %s' %mcmi)

    def get_mcmi(self, bridge):
        return self.read_file_oneline(
                 '/sys/class/net/%s/bridge/multicast_membership_interval'
                 %bridge)

    def bridge_exists(self, bridge):
        return os.path.exists('/sys/class/net/%s/bridge' %bridge)

    def is_bridge_port(self, ifacename):
        return os.path.exists('/sys/class/net/%s/brport' %ifacename)

    def bridge_port_exists(self, bridge, bridgeportname):
        try:
            return os.path.exists('/sys/class/net/%s/brif/%s'
                                  %(bridge, bridgeportname))
        except Exception:
            return False
   
    def get_bridge_ports(self, bridgename):
        try:
            return os.listdir('/sys/class/net/%s/brif/' %bridgename)
        except:
            return []



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

from utilsbase import *
from ifupdown.iface import *
from cache import *
import re

class mstpctlutil(utilsBase):
    """ This class contains helper methods to interact with mstpd using
    mstputils commands """

    _cache_fill_done = False

    _bridgeattrmap = {'bridgeid' : 'bridge-id',
                     'maxage' : 'max-age',
                     'fdelay' : 'forward-delay',
                     'txholdcount' : 'tx-hold-count',
                     'maxhops' : 'max-hops',
                     'ageing' : 'ageing-time',
                     'hello' : 'hello-time',
                     'forcevers' : 'force-protocol-version'}

    _bridgeportattrmap = {'portadminedge' : 'admin-edge-port',
                     'portp2p' : 'admin-point-to-point',
                     'portrestrrole' : 'restricted-role',
                     'portrestrtcn' : 'restricted-TCN',
                     'bpduguard' : 'bpdu-guard-port',
                     'portautoedge' : 'auto-edge-port',
                     'portnetwork' : 'network-port',
                     'portbpdufilter' : 'bpdufilter-port'}

    def __init__(self, *args, **kargs):
        utilsBase.__init__(self, *args, **kargs)

    def is_mstpd_running(self):
        try:
            self.exec_command('/bin/pidof mstpd')
        except:
            return False
        else:
            return True

    def get_bridgeport_attr(self, bridgename, portname, attrname):
        try:
            return self.subprocess_check_output(['/sbin/mstpctl',
                       'showportdetail', '%s' %bridgename, '%s' %portname,
                       self._bridgeportattrmap[attrname]]).strip('\n')
        except Exception as e:
            pass
        return None

    def get_bridgeport_attrs(self, bridgename, portname):
        bridgeattrs = {}
        try:
            bridgeattrs = dict((k, self.get_bridgeport_attr(bridgename, v))
                                 for k, v in self._bridgeattrmap.items())
            bridgeattrs['treeprio'] = int(bridgeattrs.get('bridgeid',
                                     '').split('.')[0], base=16) * 4096
        except Exception as e:
            self.logger.warn(str(e))
            pass
        return bridgeattrs

    def set_bridgeport_attrs(self, bridgename, bridgeportname, attrdict,
                             check=True):
        for k, v in attrdict.iteritems():
            if not v:
                continue
            try:
                self.set_bridgeport_attr(self, bridgename, bridgeportname,
                        k, v, check)
            except Exception as e:
                self.logger.warn(str(e))

    def set_bridgeport_attr(self, bridgename, bridgeportname, attrname,
                            attrvalue, check=True):
        if check:
            attrvalue_curr = self.get_bridgeport_attr(bridgename,
                                    bridgeportname, attrname)
            if attrvalue_curr and attrvalue_curr == attrvalue:
                return
        if attrname == 'treeportcost' or attrname == 'treeportprio':
            self.subprocess_check_output(['/sbin/mstpctl', 'set%s' %attrname,
                  '%s' %bridgename, '%s' %bridgeportname, '0', '%s' %attrvalue])
        else:
            self.subprocess_check_output(['/sbin/mstpctl', 'set%s' %attrname,
                  '%s' %bridgename, '%s' %bridgeportname, '%s' %attrvalue])

    def get_bridge_attrs(self, bridgename):
        bridgeattrs = {}
        try:
            bridgeattrs = dict((k, self.get_bridge_attr(bridgename, k))
                                 for k in self._bridgeattrmap.keys())
            bridgeattrs['treeprio'] = '%d' %(int(bridgeattrs.get('bridgeid',
                                     '').split('.')[0], base=16) * 4096)
            del bridgeattrs['bridgeid']
        except Exception as e:
            self.logger.debug(bridgeattrs)
            self.logger.debug(str(e))
            pass
        return bridgeattrs

    def get_bridge_attr(self, bridgename, attrname):
        try:
            return self.subprocess_check_output(['/sbin/mstpctl',
                       'showbridge', '%s' %bridgename,
                       self._bridgeattrmap[attrname]]).strip('\n')
        except Exception as e:
            pass
        return None

    def set_bridge_attr(self, bridgename, attrname, attrvalue, check=True):

        if check:
            attrvalue_curr = self.get_bridge_attr(bridgename, attrname)
            if attrvalue_curr and attrvalue_curr == attrvalue:
                return
        if attrname == 'treeprio':
            self.subprocess_check_call(['/sbin/mstpctl', 'set%s' %attrname,
                        '%s' %bridgename, '0',  '%s' %attrvalue])
        else:
            self.subprocess_check_call(['/sbin/mstpctl', 'set%s' %attrname,
                        '%s' %bridgename, '%s' %attrvalue])

    def set_bridge_attrs(self, bridgename, attrdict, check=True):
        for k, v in attrdict.iteritems():
            if not v:
                continue
            try:
                self.set_bridge_attr(bridgename, k, v, check)
            except Exception as e:
                self.logger.warn('%s: %s' %(bridgename, str(e)))
                pass

    def get_bridge_treeprio(self, bridgename):
        try:
            bridgeid = subprocess.check_output(['/sbin/mstpctl',
                       'showbridge', '%s' %bridgename,
                       self._bridgeattrmap['bridgeid']]).strip('\n')
            return '%d' %(int(bridgeid.split('.')[0], base=16) * 4096)
        except:
            pass
        return None

    def set_bridge_treeprio(self, bridgename, attrvalue, check=True):
        if check:
            attrvalue_curr = self.get_bridge_treeprio(bridgename)
            if attrvalue_curr and attrvalue_curr == attrvalue:
                return
        self.subprocess_check_output(['/sbin/mstpctl', 'settreeprio',
                        '%s' %bridgename, '0',  '%s' %attrvalue])

    def showbridge(self, bridgename=None):
        if bridgename:
            return self.exec_command('/sbin/mstpctl showbridge %s' %bridgename)
        else:
            return self.exec_command('/sbin/mstpctl showbridge')

    def showportdetail(self, bridgename):
        return self.exec_command('/sbin/mstpctl showportdetail %s' %bridgename)

    def mstpbridge_exists(self, bridgename):
        try:
            subprocess.check_call('mstpctl showbridge %s' %bridgename)
            return True
        except:
            return False



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

from sets import Set
from ifupdown.iface import *
from ifupdownaddons.modulebase import moduleBase
from ifupdownaddons.bridgeutils import brctl
from ifupdownaddons.iproute2 import iproute2
from ifupdownaddons.mstpctlutil import mstpctlutil
import traceback

class mstpctlFlags:
    PORT_PROCESSED = 0x1

class mstpctl(moduleBase):
    """  ifupdown2 addon module to configure mstp attributes """

    _modinfo = {'mhelp' : 'mstp configuration module for bridges',
                'attrs' : {
                   'mstpctl-ports' :
                        {'help' : 'mstp ports',
                         'compat' : True},
                   'mstpctl-stp' :
                        {'help': 'bridge stp yes/no',
                         'compat' : True,
                         'default' : 'no'},
                   'mstpctl-treeprio' :
                        {'help': 'tree priority',
                         'default' : '32768',
                         'validrange' : ['0', '65535'],
                         'required' : False,
                         'example' : ['mstpctl-treeprio 32768']},
                   'mstpctl-ageing' :
                        {'help': 'ageing time',
                         'default' : '300',
                         'required' : False,
                         'example' : ['mstpctl-ageing 300']},
                    'mstpctl-maxage' :
                        { 'help' : 'max message age',
                          'default' : '20',
                          'required' : False,
                          'example' : ['mstpctl-maxage 20']},
                    'mstpctl-fdelay' :
                        { 'help' : 'set forwarding delay',
                          'default' : '15',
                          'required' : False,
                          'example' : ['mstpctl-fdelay 15']},
                    'mstpctl-maxhops' :
                        { 'help' : 'bridge max hops',
                          'default' : '15',
                          'required' : False,
                          'example' : ['mstpctl-maxhops 15']},
                    'mstpctl-txholdcount' :
                        { 'help' : 'bridge transmit holdcount',
                          'default' : '6',
                          'required' : False,
                          'example' : ['mstpctl-txholdcount 6']},
                    'mstpctl-forcevers' :
                        { 'help' : 'bridge force stp version',
                          'default' : 'rstp',
                          'required' : False,
                          'example' : ['mstpctl-forcevers rstp']},
                    'mstpctl-portpathcost' :
                        { 'help' : 'bridge port path cost',
                          'default' : '0',
                          'required' : False,
                          'example' : ['mstpctl-portpathcost swp1=0 swp2=1']},
                    'mstpctl-portp2p' :
                        { 'help' : 'bridge port p2p detection mode',
                          'default' : 'auto',
                          'validvals' : ['yes', 'no', 'auto'],
                          'required' : False,
                          'example' : ['mstpctl-portp2p swp1=no swp2=no']},
                    'mstpctl-portrestrrole' :
                        { 'help' :
                          'enable/disable port ability to take root role of the port',
                          'default' : 'no',
                          'validvals' : ['yes', 'no'],
                          'required' : False,
                          'example' : ['mstpctl-portrestrrole swp1=no swp2=no']},
                    'mstpctl-portrestrtcn' :
                        { 'help' :
                          'enable/disable port ability to propagate received topology change notification of the port',
                          'default' : 'no',
                          'validvals' : ['yes', 'no'],
                          'required' : False,
                          'example' : ['mstpctl-portrestrtcn swp1=no swp2=no']},
                    'mstpctl-bpduguard' :
                        { 'help' :
                          'enable/disable bpduguard',
                          'default' : 'no',
                          'validvals' : ['yes', 'no'],
                          'required' : False,
                          'example' : ['mstpctl-bpduguard swp1=no swp2=no']},
                    'mstpctl-treeportprio' : 
                        { 'help' :
                          'port priority for MSTI instance',
                          'default' : '128',
                          'validrange' : ['0', '240'],
                          'required' : False,
                          'example' : ['mstpctl-treeportprio swp1=128 swp2=128']},
                    'mstpctl-hello' :
                        { 'help' : 'set hello time',
                          'default' : '2',
                          'required' : False,
                          'example' : ['mstpctl-hello 2']},
                    'mstpctl-portnetwork' : 
                        { 'help' : 'enable/disable bridge assurance capability for a port',
                          'validvals' : ['yes', 'no'],
                          'default' : 'no',
                          'required' : False,
                          'example' : ['mstpctl-portnetwork swp1=no swp2=no']},
                    'mstpctl-portadminedge' : 
                        { 'help' : 'enable/disable initial edge state of the port',
                          'validvals' : ['yes', 'no'],
                          'default' : 'no',
                          'required' : False,
                          'example' : ['mstpctl-portadminedge swp1=no swp2=no']},
                    'mstpctl-portautoedge' : 
                        { 'help' : 'enable/disable auto transition to/from edge state of the port',
                          'validvals' : ['yes', 'no'],
                          'default' : 'yes',
                          'required' : False,
                          'example' : ['mstpctl-portautoedge swp1=yes swp2=yes']},
                    'mstpctl-treeportcost' : 
                        { 'help' : 'port tree cost',
                          'required' : False},
                    'mstpctl-portbpdufilter' : 
                        { 'help' : 'enable/disable bpdu filter on a port. ' +
                                'syntax varies when defined under a bridge ' +
                                'vs under a port',
                          'validvals' : ['yes', 'no'],
                          'default' : 'no',
                          'required' : False,
                          'example' : ['under a bridge: mstpctl-portbpdufilter swp1=no swp2=no',
                                       'under a port: mstpctl-portbpdufilter yes']},
                        }}

    # Maps mstp bridge attribute names to corresponding mstpctl commands
    # XXX: This can be encoded in the modules dict above
    _attrs_map = OrderedDict([('mstpctl-treeprio' , 'treeprio'),
                  ('mstpctl-ageing' , 'ageing'),
                  ('mstpctl-maxage' , 'maxage'),
                  ('mstpctl-fdelay' , 'fdelay'),
                  ('mstpctl-maxhops' , 'maxhops'),
                  ('mstpctl-txholdcount' , 'txholdcount'),
                  ('mstpctl-forcevers', 'forcevers'),
                  ('mstpctl-hello' , 'hello')])

    # Maps mstp port attribute names to corresponding mstpctl commands
    # XXX: This can be encoded in the modules dict above
    _port_attrs_map = {'mstpctl-portpathcost' : 'portpathcost',
                 'mstpctl-portadminedge' : 'portadminedge',
                 'mstpctl-portautoedge' : 'portautoedge' ,
                 'mstpctl-portp2p' : 'portp2p',
                 'mstpctl-portrestrrole' : 'portrestrrole',
                 'mstpctl-portrestrtcn' : 'portrestrtcn',
                 'mstpctl-bpduguard' : 'bpduguard',
                 'mstpctl-treeportprio' : 'treeportprio',
                 'mstpctl-treeportcost' : 'treeportcost',
                 'mstpctl-portnetwork' : 'portnetwork',
                 'mstpctl-portbpdufilter' : 'portbpdufilter'}

    def __init__(self, *args, **kargs):
        moduleBase.__init__(self, *args, **kargs)
        self.ipcmd = None
        self.name = self.__class__.__name__
        self.brctlcmd = None
        self.mstpctlcmd = None

    def _is_bridge(self, ifaceobj):
        if (ifaceobj.get_attr_value_first('mstpctl-ports') or
                ifaceobj.get_attr_value_first('bridge-ports')):
            return True
        return False

    def _is_bridge_port(self, ifaceobj):
        if self.brctlcmd.is_bridge_port(ifaceobj.name):
            return True
        return False

    def get_dependent_ifacenames(self, ifaceobj, ifacenames_all=None):
        if not self._is_bridge(ifaceobj):
            return None
        return self.parse_port_list(ifaceobj.get_attr_value_first(
                                    'mstpctl-ports'), ifacenames_all)

    def get_dependent_ifacenames_running(self, ifaceobj):
        self._init_command_handlers()
        if (self.brctlcmd.bridge_exists(ifaceobj.name) and
                not self.mstpctlcmd.mstpbridge_exists(ifaceobj.name)):
            return None
        return self.brctlcmd.get_bridge_ports(ifaceobj.name)

    def _get_bridge_port_list(self, ifaceobj):

        # port list is also available in the previously
        # parsed dependent list. Use that if available, instead
        # of parsing port expr again
        port_list = ifaceobj.lowerifaces
        if port_list:
            return port_list
        ports = ifaceobj.get_attr_value_first('mstpctl-ports')
        if ports:
            return self.parse_port_list(ports)
        else:
            return None

    def _ports_enable_disable_ipv6(self, ports, enable='1'):
        for p in ports:
            try:
                self.write_file('/proc/sys/net/ipv6/conf/%s' %p +
                                '/disable_ipv6', enable)
            except Exception as e:
                self.logger.info(str(e))
                pass

    def _add_ports(self, ifaceobj):
        bridgeports = self._get_bridge_port_list(ifaceobj)

        runningbridgeports = []
        # Delete active ports not in the new port list
        if not self.PERFMODE:
            runningbridgeports = self.brctlcmd.get_bridge_ports(ifaceobj.name)
            if runningbridgeports:
                [self.ipcmd.link_set(bport, 'nomaster')
                    for bport in runningbridgeports
                        if not bridgeports or bport not in bridgeports]
            else:
                runningbridgeports = []
        if not bridgeports:
            return
        err = 0
        for bridgeport in Set(bridgeports).difference(Set(runningbridgeports)):
            try:
                if not self.DRYRUN and not self.ipcmd.link_exists(bridgeport):
                    self.log_warn('%s: bridge port %s does not exist'
                            %(ifaceobj.name, bridgeport))
                    err += 1
                    continue
                self.ipcmd.link_set(bridgeport, 'master', ifaceobj.name)
                self.ipcmd.addr_flush(bridgeport)
            except Exception as e:
                self.log_error(str(e))

        if err:
            self.log_error('error configuring bridge (missing ports)')

    def _apply_bridge_settings(self, ifaceobj):
        check = False if self.PERFMODE else True
        try:
            # set bridge attributes
            for attrname, dstattrname in self._attrs_map.items():
                try:
                    v = ifaceobj.get_attr_value_first(attrname)
                    if not v:
                       continue
                    if attrname == 'mstpctl-treeprio':
                       self.mstpctlcmd.set_bridge_treeprio(ifaceobj.name,
                                v, check)
                    else:
                       self.mstpctlcmd.set_bridge_attr(ifaceobj.name,
                                dstattrname, v, check)
                except Exception as e:
                    self.logger.warn('%s' %str(e))
                    pass

            # set bridge port attributes
            for attrname, dstattrname in self._port_attrs_map.items():
                attrval = ifaceobj.get_attr_value_first(attrname)
                if not attrval:
                    continue
                portlist = self.parse_port_list(attrval)
                if not portlist:
                    self.log_warn('%s: error parsing \'%s %s\''
                         %(ifaceobj.name, attrname, attrval))
                    continue
                for p in portlist:
                    try:
                        (port, val) = p.split('=')
                        self.mstpctlcmd.set_bridgeport_attr(ifaceobj.name,
                                port, dstattrname, val, check)
                    except Exception as e:
                        self.log_warn('%s: error setting %s (%s)'
                                %(ifaceobj.name, attrname, str(e)))
        except Exception as e:
            self.log_warn(str(e))
            pass

    def _apply_bridge_port_settings(self, ifaceobj, bridgename=None,
                                    bridgeifaceobj=None, stp_on=True,
                                    mstpd_running=True):
        check = False if self.PERFMODE else True
        if not bridgename and bridgeifaceobj:
            bridgename = bridgeifaceobj.name
        # set bridge port attributes
        for attrname, dstattrname in self._port_attrs_map.items():
            attrval = ifaceobj.get_attr_value_first(attrname)
            if not attrval:
               #if bridgeifaceobj:
               #   # If bridge object available, check if the bridge
               #   # has the attribute set, in which case,
               #   # inherit it from the bridge
               #   attrval = bridgeifaceobj.get_attr_value_first(attrname)
               #   if not attrval:
               #      continue
               #else:
               continue
            if not stp_on:
               self.logger.warn('%s: cannot set %s (stp on bridge %s not on)\n'
                       %(ifaceobj.name, attrname, bridgename))
               continue
            if not mstpd_running:
               continue
            try:
               self.mstpctlcmd.set_bridgeport_attr(bridgename,
                           ifaceobj.name, dstattrname, attrval, check)
            except Exception as e:
               self.log_warn('%s: error setting %s (%s)'
                             %(ifaceobj.name, attrname, str(e)))

    def _apply_bridge_port_settings_all(self, ifaceobj,
                                        ifaceobj_getfunc=None):
        self.logger.info('%s: applying mstp configuration '
                          %ifaceobj.name + 'specific to ports')
        # Query running bridge ports. and only apply attributes on them
        bridgeports = self.brctlcmd.get_bridge_ports(ifaceobj.name)
        if not bridgeports:
           self.logger.debug('%s: cannot find bridgeports' %ifaceobj.name)
           return
        for bport in bridgeports:
            self.logger.info('%s: processing mstp config for port %s'
                             %(ifaceobj.name, bport))
            if not self.ipcmd.link_exists(bport):
               continue
            bportifaceobjlist = ifaceobj_getfunc(bport)
            if not bportifaceobjlist:
               continue
            for bportifaceobj in bportifaceobjlist:
                # Dont process bridge port if it already has been processed
                if (bportifaceobj.module_flags.get(self.name,0x0) & \
                    mstpctlFlags.PORT_PROCESSED):
                    continue
                try:
                    self._apply_bridge_port_settings(bportifaceobj, 
                                            ifaceobj.name, ifaceobj)
                except Exception as e:
                    self.log_warn(str(e))

    def _up(self, ifaceobj, ifaceobj_getfunc=None):
        # Check if bridge port
        bridgename = self.ipcmd.bridge_port_get_bridge_name(ifaceobj.name)
        if bridgename:
            mstpd_running = (True if self.mstpctlcmd.is_mstpd_running()
                             else False)
            stp_on = (True if self.read_file_oneline(
                      '/sys/class/net/%s/bridge/stp_state'
                      %bridgename) == '2' else False)
            self._apply_bridge_port_settings(ifaceobj, bridgename, None,
                                             stp_on, mstpd_running)
            ifaceobj.module_flags[self.name] = ifaceobj.module_flags.setdefault(self.name,0) | \
                                               mstpctlFlags.PORT_PROCESSED
            return
        if not self._is_bridge(ifaceobj):
            return
        stp = None
        try:
            porterr = False
            porterrstr = ''
            if ifaceobj.get_attr_value_first('mstpctl-ports'):
                # If bridge ports specified with mstpctl attr, create the
                # bridge and also add its ports
                self.ipcmd.batch_start()
                if not self.PERFMODE:
                    if not self.ipcmd.link_exists(ifaceobj.name):
                        self.ipcmd.link_create(ifaceobj.name, 'bridge')
                else:
                    self.ipcmd.link_create(ifaceobj.name, 'bridge')
                try:
                    self._add_ports(ifaceobj)
                except Exception as e:
                    porterr = True
                    porterrstr = str(e)
                    pass
                finally:
                    self.ipcmd.batch_commit()
            running_ports = self.brctlcmd.get_bridge_ports(ifaceobj.name)
            if running_ports:
                # disable ipv6 for ports that were added to bridge
                self._ports_enable_disable_ipv6(running_ports, '1')

            stp = ifaceobj.get_attr_value_first('mstpctl-stp')
            if stp:
               self.set_iface_attr(ifaceobj, 'mstpctl-stp',
                                    self.brctlcmd.set_stp)
            else:
               stp = self.brctlcmd.get_stp(ifaceobj.name)
            if (self.mstpctlcmd.is_mstpd_running() and
                    (stp == 'yes' or stp == 'on')):
                self._apply_bridge_settings(ifaceobj)
                self._apply_bridge_port_settings_all(ifaceobj,
                            ifaceobj_getfunc=ifaceobj_getfunc)
        except Exception as e:
            self.log_error(str(e))
        if porterr:
            raise Exception(porterrstr)

    def _down(self, ifaceobj, ifaceobj_getfunc=None):
        if not self._is_bridge(ifaceobj):
            return
        try:
            if ifaceobj.get_attr_value_first('mstpctl-ports'):
                # If bridge ports specified with mstpctl attr, delete the
                # bridge
                ports = self.brctlcmd.get_bridge_ports(ifaceobj.name)
                if ports:
                    self._ports_enable_disable_ipv6(ports, '0')
                self.brctlcmd.delete_bridge(ifaceobj.name)
        except Exception as e:
            self.log_error(str(e))

    def _query_running_attrs(self, ifaceobjrunning):
        bridgeattrdict = {}

        tmpbridgeattrdict = self.mstpctlcmd.get_bridge_attrs(ifaceobjrunning.name)
        if not tmpbridgeattrdict:
            return bridgeattrdict

        for k,v in tmpbridgeattrdict.items():
            if k == 'stp' or not v:
                continue
            if k == 'ports':
                ports = v.keys()
                continue
            attrname = 'mstpctl-' + k
            if v and v != self.get_mod_subattr(attrname, 'default'):
                bridgeattrdict[attrname] = [v]

        ports = self.brctlcmd.get_bridge_ports(ifaceobjrunning.name)
        if ports:
            portconfig = {'mstpctl-portnetwork' : '',
                          'mstpctl-portpathcost' : '',
                          'mstpctl-portadminedge' : '',
                          'mstpctl-portautoedge' : '',
                          'mstpctl-portp2p' : '',
                          'mstpctl-portrestrrole' : '',
                          'mstpctl-portrestrtcn' : '',
                          'mstpctl-bpduguard' : '',
                          'mstpctl-treeportprio' : '',
                          'mstpctl-treeportcost' : ''}

            for p in ports:
                v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                            p, 'portnetwork')
                if v and v != 'no':
                    portconfig['mstpctl-portnetwork'] += ' %s=%s' %(p, v)

                # XXX: Can we really get path cost of a port ???
                #v = self.mstpctlcmd.get_portpathcost(ifaceobjrunning.name, p)
                #if v and v != self.get_mod_subattr('mstpctl-portpathcost',
                #                                   'default'):
                #    portconfig['mstpctl-portpathcost'] += ' %s=%s' %(p, v)

                v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                            p, 'portadminedge')
                if v and v != 'no':
                    portconfig['mstpctl-portadminedge'] += ' %s=%s' %(p, v)

                v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                            p, 'portp2p')
                if v and v != 'no':
                    portconfig['mstpctl-portp2p'] += ' %s=%s' %(p, v)

                v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                            p, 'portrestrrole')
                if v and v != 'no':
                    portconfig['mstpctl-portrestrrole'] += ' %s=%s' %(p, v)

                v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                            p, 'portrestrtcn')
                if v and v != 'no':
                    portconfig['mstpctl-portrestrtcn'] += ' %s=%s' %(p, v)

                v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                            p, 'bpduguard')
                if v and v != 'no':
                    portconfig['mstpctl-bpduguard'] += ' %s=%s' %(p, v)

                # XXX: Can we really get path cost of a port ???
                #v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                #            p, 'treeprio')
                #if v and v != self.get_mod_subattr('mstpctl-treeportprio',
                #                                   'default'):
                #    portconfig['mstpctl-treeportprio'] += ' %s=%s' %(p, v)

                #v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
                #            p, 'treecost')
                #if v and v != self.get_mod_subattr('mstpctl-treeportcost',
                #                                   'default'):
                #    portconfig['mstpctl-treeportcost'] += ' %s=%s' %(p, v)

            bridgeattrdict.update({k : [v] for k, v in portconfig.items()
                                    if v})
        return bridgeattrdict

    def _query_check_bridge(self, ifaceobj, ifaceobjcurr):
        # list of attributes that are not supported currently
        blacklistedattrs = ['mstpctl-portpathcost',
                'mstpctl-treeportprio', 'mstpctl-treeportcost']
        if not self.brctlcmd.bridge_exists(ifaceobj.name):
            self.logger.debug('bridge %s does not exist' %ifaceobj.name)
            return
        ifaceattrs = self.dict_key_subset(ifaceobj.config,
                                          self.get_mod_attrs())
        if not ifaceattrs:
            return
        runningattrs = self.mstpctlcmd.get_bridge_attrs(ifaceobj.name)
        if not runningattrs:
            runningattrs = {}
        for k in ifaceattrs:
            # for all mstpctl options
            if k in blacklistedattrs:
                continue
            # get the corresponding ifaceobj attr
            v = ifaceobj.get_attr_value_first(k)
            if not v:
                continue

            # Get the running attribute
            rv = runningattrs.get(k[8:])
            if k == 'mstpctl-stp':
                # special case stp compare because it may
                # contain more than one valid values
                stp_on_vals = ['on', 'yes']
                stp_off_vals = ['off']
                rv = self.brctlcmd.get_stp(ifaceobj.name)
                if ((v in stp_on_vals and rv in stp_on_vals) or
                    (v in stp_off_vals and rv in stp_off_vals)):
                    ifaceobjcurr.update_config_with_status('mstpctl-stp', v, 0)
                else:
                    ifaceobjcurr.update_config_with_status('mstpctl-stp', v, 1)
                continue

            if k == 'mstpctl-ports':
                # special case ports because it can contain regex or glob
                # XXX: We get all info from mstputils, which means if
                # mstpd is down, we will not be returning any bridge bridgeports
                running_port_list = self.brctlcmd.get_bridge_ports(ifaceobj.name)
                bridge_port_list = self._get_bridge_port_list(ifaceobj)
                if not running_port_list and not bridge_port_list:
                    continue
                portliststatus = 1
                if running_port_list and bridge_port_list:
                    difference = Set(running_port_list).symmetric_difference(
                                                        Set(bridge_port_list))
                    if not difference:
                        portliststatus = 0
                ifaceobjcurr.update_config_with_status('mstpctl-ports',
                    ' '.join(running_port_list)
                    if running_port_list else '', portliststatus)
            elif k[:12] == 'mstpctl-port' or k == 'mstpctl-bpduguard':
                # Now, look at port attributes
                # derive the mstpctlcmd attr name
                #mstpctlcmdattrname = k[12:] if k[:12] == 'mstpctl-port' else k[8:]
                mstpctlcmdattrname = k[8:]

                # for port attributes, the attributes are in a list
                # <portname>=<portattrvalue>
                status = 0
                currstr = ''
                vlist = self.parse_port_list(v)
                if not vlist:
                    continue
                for vlistitem in vlist:
                    try:
                        (p, v) = vlistitem.split('=')
                        currv = self.mstpctlcmd.get_bridgeport_attr(
                                        ifaceobj.name, p, mstpctlcmdattrname)
                        if currv:
                            currstr += ' %s=%s' %(p, currv)
                        else:
                            currstr += ' %s=%s' %(p, 'None')
                        if currv != v:
                            status = 1
                    except Exception as e:
                        self.log_warn(str(e))
                        pass
                ifaceobjcurr.update_config_with_status(k, currstr, status)
            elif not rv:
                ifaceobjcurr.update_config_with_status(k, '', 1)
            elif v != rv:
                ifaceobjcurr.update_config_with_status(k, rv, 1)
            else:
                ifaceobjcurr.update_config_with_status(k, rv, 0)

    def _query_check_bridge_port(self, ifaceobj, ifaceobjcurr):
        if not self.ipcmd.link_exists(ifaceobj.name):
            #self.logger.debug('bridge port %s does not exist' %ifaceobj.name)
            ifaceobjcurr.status = ifaceStatus.NOTFOUND
            return
        # Check if this is a bridge port
        if not self._is_bridge_port(ifaceobj):
            # mark all the bridge attributes as error
            ifaceobjcurr.check_n_update_config_with_status_many(ifaceobj,
                            self._port_attrs_map.keys(), 0)
            return
        bridgename = self.ipcmd.bridge_port_get_bridge_name(ifaceobj.name)
        # list of attributes that are not supported currently
        blacklistedattrs = ['mstpctl-portpathcost',
                'mstpctl-treeportprio', 'mstpctl-treeportcost']
        ifaceattrs = self.dict_key_subset(ifaceobj.config,
                                          self._port_attrs_map.keys())
        if not ifaceattrs:
            return
        runningattrs = self.mstpctlcmd.get_bridge_attrs(ifaceobj.name)
        if not runningattrs:
            runningattrs = {}
        for k in ifaceattrs:
            # for all mstpctl options
            # get the corresponding ifaceobj attr
            v = ifaceobj.get_attr_value_first(k)
            if not v or k in blacklistedattrs:
                ifaceobjcurr.update_config_with_status(k, v, -1)
                continue
            currv = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                             ifaceobj.name, self._port_attrs_map.get(k))
            if currv:
                if currv != v:
                    ifaceobjcurr.update_config_with_status(k, currv, 1)
                else:
                    ifaceobjcurr.update_config_with_status(k, currv, 0)
            else:
                ifaceobjcurr.update_config_with_status(k, None, 1)

    def _query_check(self, ifaceobj, ifaceobjcurr, ifaceobj_getfunc=None):
        if self._is_bridge(ifaceobj):
            self._query_check_bridge(ifaceobj, ifaceobjcurr)
        else:
            self._query_check_bridge_port(ifaceobj, ifaceobjcurr)

    def _query_running_bridge_port(self, ifaceobjrunning):
        bridgename = self.ipcmd.bridge_port_get_bridge_name(
                                ifaceobjrunning.name)
        if not bridgename:
            self.logger.warn('%s: unable to determine bridgename'
                             %ifaceobjrunning.name)
            return
        if self.brctlcmd.get_stp(bridgename) == 'no':
           # This bridge does not run stp, return
           return
        # if userspace stp not set, return
        if self.sysctl_get('net.bridge.bridge-stp-user-space') != '1':
           return
        v = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                                                ifaceobjrunning.name,
                                                'portnetwork')
        if v and v != 'no':
           ifaceobjrunning.update_config('mstpctl-network', v)

        # XXX: Can we really get path cost of a port ???
        #v = self.mstpctlcmd.get_portpathcost(ifaceobjrunning.name, p)
        #if v and v != self.get_mod_subattr('mstpctl-pathcost',
        #                                   'default'):
        #   ifaceobjrunning.update_config('mstpctl-network', v)

        v = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                          ifaceobjrunning.name, 'portadminedge')
        if v and v != 'no':
           ifaceobjrunning.update_config('mstpctl-portadminedge', v)

        v = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                                       ifaceobjrunning.name,'portp2p')
        if v and v != 'auto':
           ifaceobjrunning.update_config('mstpctl-portp2p', v)

        v = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                        ifaceobjrunning.name, 'portrestrrole')
        if v and v != 'no':
           ifaceobjrunning.update_config('mstpctl-portrestrrole', v)

        v = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                            ifaceobjrunning.name, 'restrtcn')
        if v and v != 'no':
           ifaceobjrunning.update_config('mstpctl-portrestrtcn', v)

        v = self.mstpctlcmd.get_bridgeport_attr(bridgename,
                            ifaceobjrunning.name, 'bpduguard')
        if v and v != 'no':
           ifaceobjrunning.update_config('mstpctl-bpduguard', v)

        # XXX: Can we really get path cost of a port ???
        #v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
        #            p, 'treeprio')
        #if v and v != self.get_mod_subattr('mstpctl-treeportprio',
        #                                   'default'):
        #    portconfig['mstpctl-treeportprio'] += ' %s=%s' %(p, v)

        #v = self.mstpctlcmd.get_bridgeport_attr(ifaceobjrunning.name,
        #               p, 'treecost')
        #if v and v != self.get_mod_subattr('mstpctl-treeportcost',
        #                                   'default'):
        #    portconfig['mstpctl-treeportcost'] += ' %s=%s' %(p, v)

    def _query_running_bridge(self, ifaceobjrunning):
        if self.brctlcmd.get_stp(ifaceobjrunning.name) == 'no':
           # This bridge does not run stp, return
           return
        # if userspace stp not set, return
        if self.sysctl_get('net.bridge.bridge-stp-user-space') != '1':
           return
        # Check if mstp really knows about this bridge
        if not self.mstpctlcmd.mstpbridge_exists(ifaceobjrunning.name):
            return
        ifaceobjrunning.update_config_dict(self._query_running_attrs(
                                           ifaceobjrunning))

    def _query_running(self, ifaceobjrunning, **extra_args):
        if self.brctlcmd.bridge_exists(ifaceobjrunning.name):
            self._query_running_bridge(ifaceobjrunning)
        elif self.brctlcmd.is_bridge_port(ifaceobjrunning.name):
            self._query_running_bridge_port(ifaceobjrunning)

    _run_ops = {'pre-up' : _up,
               'post-down' : _down,
               'query-checkcurr' : _query_check,
               'query-running' : _query_running}

    def get_ops(self):
        """ returns list of ops supported by this module """
        return self._run_ops.keys()

    def _init_command_handlers(self):
        flags = self.get_flags()
        if not self.ipcmd:
            self.ipcmd = iproute2(**flags)
        if not self.brctlcmd:
            self.brctlcmd = brctl(**flags)
        if not self.mstpctlcmd:
            self.mstpctlcmd = mstpctlutil(**flags)

    def run(self, ifaceobj, operation, query_ifaceobj=None,
            ifaceobj_getfunc=None, **extra_args):
        """ run mstp configuration on the interface object passed as argument

        Args:
            **ifaceobj** (object): iface object

            **operation** (str): any of 'pre-up', 'post-down', 'query-checkcurr',
                                 'query-running'
        Kwargs:
            **query_ifaceobj** (object): query check ifaceobject. This is only
                valid when op is 'query-checkcurr'. It is an object same as
                ifaceobj, but contains running attribute values and its config
                status. The modules can use it to return queried running state
                of interfaces. status is success if the running state is same
                as user required state in ifaceobj. error otherwise.
        """
        if ifaceobj.type == ifaceType.BRIDGE_VLAN:
           return
        op_handler = self._run_ops.get(operation)
        if not op_handler:
           return
        self._init_command_handlers()
        if operation == 'query-checkcurr':
            op_handler(self, ifaceobj, query_ifaceobj,
                       ifaceobj_getfunc=ifaceobj_getfunc)
        else:
            op_handler(self, ifaceobj, ifaceobj_getfunc=ifaceobj_getfunc)



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

import subprocess
import ifupdownaddons

class usercmds(ifupdownaddons.modulebase.moduleBase):
    """  ifupdown2 addon module to configure user specified commands """

    _modinfo = {'mhelp' : 'user commands for interfaces',
                'attrs' : {
                   'pre-up' :
                        {'help' : 'run command before bringing the interface up'},
                   'up' :
                        {'help' : 'run command at interface bring up'},
                   'post-up' :
                        {'help' : 'run command after interface bring up'},
                   'pre-down' :
                        {'help' : 'run command before bringing the interface down'},
                   'down' :
                        {'help' : 'run command at interface down'},
                   'post-down' :
                        {'help' : 'run command after bringing interface down'}}}

    def _exec_user_cmd(self, cmd):
        """ exec's commands using subprocess Popen

        special wrapper using use closefds=True and shell=True
        for user commands
        """

        cmd_returncode = 0
        try:
            self.logger.info('executing %s' %cmd)
            if self.DRYRUN:
                return
            ch = subprocess.Popen(cmd,
                    stdout=subprocess.PIPE,
                    shell=True,
                    stderr=subprocess.STDOUT,
                    close_fds=True)
            cmd_returncode = ch.wait()
            cmdout = ch.communicate()[0]
        except Exception as e:
            raise Exception('failed to execute cmd \'%s\' (%s)'
                            %(cmd, str(e)))
        if cmd_returncode != 0:
            raise Exception(cmdout)
        return cmdout

    def _run_command(self, ifaceobj, op):
        cmd_list = ifaceobj.get_attr_value(op)
        if cmd_list:
            for cmd in cmd_list:
                self.logger.info('executing cmd \'%s\'' %cmd)
                try:
                    self._exec_user_cmd(cmd)
                except Exception as e:
                    if not self.ignore_error(str(e)):
                        self.logger.warn('%s: %s cmd \'%s\' failed (%s)'
                                %(ifaceobj.name, op, cmd, str(e).strip('\n')))
                    pass

    _run_ops = {'pre-up' : _run_command,
               'pre-down' : _run_command,
               'up' : _run_command,
               'post-up' : _run_command,
               'down' : _run_command,
               'post-down' : _run_command}

    def get_ops(self):
        """ returns list of ops supported by this module """
        return self._run_ops.keys()

    def run(self, ifaceobj, operation, query_ifaceobj=None, **extra_args):
        """ run user commands

        Args:
            **ifaceobj** (object): iface object

            **operation** (str): list of ops

        Kwargs:
            **query_ifaceobj** (object): query check ifaceobject. This is only
                valid when op is 'query-checkcurr'. It is an object same as
                ifaceobj, but contains running attribute values and its config
                status. The modules can use it to return queried running state
                of interfaces. status is success if the running state is same
                as user required state in ifaceobj. error otherwise.
        """
        op_handler = self._run_ops.get(operation)
        if not op_handler:
            return
        op_handler(self, ifaceobj, operation)



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#
# ifupdownBase --
#    base object for various ifupdown objects
#

import logging
import subprocess
import re
import os
import traceback
from iface import *
import rtnetlink_api as rtnetlink_api

class ifupdownBase(object):

    def __init__(self):
        modulename = self.__class__.__name__
        self.logger = logging.getLogger('ifupdown.' + modulename)

    def exec_command(self, cmd, cmdenv=None, nowait=False):
        cmd_returncode = 0
        cmdout = ''
        try:
            self.logger.info('Executing ' + cmd)
            if self.DRYRUN:
                return cmdout
            ch = subprocess.Popen(cmd.split(),
                    stdout=subprocess.PIPE,
                    shell=False, env=cmdenv,
                    stderr=subprocess.STDOUT,
                    close_fds=True)
            cmdout = ch.communicate()[0]
            cmd_returncode = ch.wait()
        except OSError as e:
            raise Exception('could not execute ' + cmd +
                    '(' + str(e) + ')')
        if cmd_returncode != 0:
            raise Exception('error executing cmd \'%s\'' %cmd +
                '\n(' + cmdout.strip('\n ') + ')')
        return cmdout

    def ignore_error(self, errmsg):
        if (self.FORCE == True or re.search(r'exists', errmsg,
            re.IGNORECASE | re.MULTILINE) is not None):
            return True
        return False

    def log_warn(self, str):
        if self.ignore_error(str) == False:
            if self.logger.getEffectiveLevel() == logging.DEBUG:
                traceback.print_stack()
            self.logger.warn(str)
        pass

    def log_error(self, str):
        if self.ignore_error(str) == False:
            raise
            #raise Exception(str)
        else:
            pass

    def link_exists(self, ifacename):
        return os.path.exists('/sys/class/net/%s' %ifacename)

    def link_up(self, ifacename):
        rtnetlink_api.rtnl_api.link_set(ifacename, "up")

    def link_down(self, ifacename):
        rtnetlink_api.rtnl_api.link_set(ifacename, "down")



#!/usr/bin/python

from ifupdown.iface import *
from ifupdownaddons.modulebase import moduleBase
from ifupdownaddons.iproute2 import iproute2
import ifupdown.rtnetlink_api as rtnetlink_api
import logging
from sets import Set

class vxlan(moduleBase):
    _modinfo = {'mhelp' : 'vxlan module configures vxlan interfaces.',
                'attrs' : {
                        'vxlan-id' :
                            {'help' : 'vxlan id',
                             'required' : True,
                             'example': ['vxlan-id 100']},
                        'vxlan-local-tunnelip' :
                            {'help' : 'vxlan local tunnel ip',
                             'example': ['vxlan-local-tunnelip 172.16.20.103']},
                        'vxlan-svcnodeip' :
                            {'help' : 'vxlan id',
                             'example': ['vxlan-svcnodeip 172.16.22.125']},
                        'vxlan-remoteip' :
                            {'help' : 'vxlan remote ip',
                             'example': ['vxlan-remoteip 172.16.22.127']},
                        'vxlan-learning' :
                            {'help' : 'vxlan learning on/off',
                             'example': ['vxlan-learning off'],
                             'default': 'on'},
                }}

    def __init__(self, *args, **kargs):
        moduleBase.__init__(self, *args, **kargs)
        self.ipcmd = None

    def get_dependent_ifacenames(self, ifaceobj, ifaceobjs_all=None):
        if not self._is_vxlan_device(ifaceobj):
            return None
        ifaceobj.link_kind |= ifaceLinkKind.VXLAN
        return None

    def _is_vxlan_device(self, ifaceobj):
        if ifaceobj.get_attr_value_first('vxlan-id'):
            return True
        return False

    def _up(self, ifaceobj):
        vxlanid = ifaceobj.get_attr_value_first('vxlan-id')
        if vxlanid:
            self.ipcmd.link_create_vxlan(ifaceobj.name, vxlanid,
            localtunnelip=ifaceobj.get_attr_value_first('vxlan-local-tunnelip'),
            svcnodeips=ifaceobj.get_attr_value('vxlan-svcnodeip'),
            remoteips=ifaceobj.get_attr_value('vxlan-remoteip'),
            learning=ifaceobj.get_attr_value_first('vxlan-learning'),
            ageing=ifaceobj.get_attr_value_first('vxlan-ageing'))
            if ifaceobj.addr_method == 'manual':
               rtnetlink_api.rtnl_api.link_set(ifaceobj.name, "up")

    def _down(self, ifaceobj):
        try:
            self.ipcmd.link_delete(ifaceobj.name)
        except Exception as e:
            self.log_warn(str(e))

    def _query_check_n_update(self, ifaceobjcurr, attrname, attrval,
                              running_attrval):
        if running_attrval and attrval == running_attrval:
           ifaceobjcurr.update_config_with_status(attrname, attrval, 0)
        else:
           ifaceobjcurr.update_config_with_status(attrname, running_attrval, 1)

    def _query_check_n_update_addresses(self, ifaceobjcurr, attrname,
                                        addresses, running_addresses):
        if addresses:
            for a in addresses: 
                if a in running_addresses:
                    ifaceobjcurr.update_config_with_status(attrname, a, 0)
                else:
                    ifaceobjcurr.update_config_with_status(attrname, a, 1)
            running_addresses = Set(running_addresses).difference(
                                                    Set(addresses))
        [ifaceobjcurr.update_config_with_status(attrname, a, 1)
                    for a in running_addresses]

    def _query_check(self, ifaceobj, ifaceobjcurr):
        if not self.ipcmd.link_exists(ifaceobj.name):
           return
        # Update vxlan object
        vxlanattrs = self.ipcmd.get_vxlandev_attrs(ifaceobj.name)
        if not vxlanattrs:
            ifaceobjcurr.check_n_update_config_with_status_many(ifaceobj,
                    self.get_mod_attrs(), -1)
            return
        self._query_check_n_update(ifaceobjcurr, 'vxlan-id',
                       ifaceobj.get_attr_value_first('vxlan-id'), 
                       vxlanattrs.get('vxlanid'))

        self._query_check_n_update(ifaceobjcurr, 'vxlan-local-tunnelip',
                       ifaceobj.get_attr_value_first('vxlan-local-tunnelip'), 
                       vxlanattrs.get('local'))

        self._query_check_n_update_addresses(ifaceobjcurr, 'vxlan-svcnodeip',
                       ifaceobj.get_attr_value('vxlan-svcnodeip'), 
                       vxlanattrs.get('svcnode', []))

        self._query_check_n_update_addresses(ifaceobjcurr, 'vxlan-remoteip',
                       ifaceobj.get_attr_value('vxlan-remoteip'), 
                       vxlanattrs.get('remote', []))

        learning = ifaceobj.get_attr_value_first('vxlan-learning')
        if not learning:
            learning = 'on'
        running_learning = vxlanattrs.get('learning')
        if learning == running_learning:
           ifaceobjcurr.update_config_with_status('vxlan-learning',
                                                  running_learning, 0)
        else:
           ifaceobjcurr.update_config_with_status('vxlan-learning',
                                                  running_learning, 1)

    def _query_running(self, ifaceobjrunning):
        vxlanattrs = self.ipcmd.get_vxlandev_attrs(ifaceobjrunning.name)
        if not vxlanattrs:
            return
        attrval = vxlanattrs.get('vxlanid')
        if attrval:
            ifaceobjrunning.update_config('vxlan-id', vxlanattrs.get('vxlanid'))
        attrval = vxlanattrs.get('local')
        if attrval:
            ifaceobjrunning.update_config('vxlan-local-tunnelip', attrval)
        attrval = vxlanattrs.get('svcnode')
        if attrval:
            [ifaceobjrunning.update_config('vxlan-svcnode', a)
                        for a in attrval]
        attrval = vxlanattrs.get('remote')
        if attrval:
            [ifaceobjrunning.update_config('vxlan-remoteip', a)
                        for a in attrval]
        attrval = vxlanattrs.get('learning')
        if attrval and attrval == 'on':
            ifaceobjrunning.update_config('vxlan-learning', 'on')


    _run_ops = {'pre-up' : _up,
               'post-down' : _down,
               'query-checkcurr' : _query_check,
               'query-running' : _query_running}

    def get_ops(self):
        return self._run_ops.keys()

    def _init_command_handlers(self):
        if not self.ipcmd:
            self.ipcmd = iproute2(**self.get_flags())

    def run(self, ifaceobj, operation, query_ifaceobj=None, **extra_args):
        op_handler = self._run_ops.get(operation)
        if not op_handler:
            return
        if (operation != 'query-running' and
                not self._is_vxlan_device(ifaceobj)):
            return
        self._init_command_handlers()
        if operation == 'query-checkcurr':
            op_handler(self, ifaceobj, query_ifaceobj)
        else:
            op_handler(self, ifaceobj)



#!/usr/bin/env python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
#
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#
#

from os import getpid
from socket import AF_UNSPEC
from socket import AF_BRIDGE
from iff import IFF_UP
from rtnetlink import *
import os
import ifupdownmain

class rtnetlinkApi(RtNetlink):

    bind_done = False

    def __init__(self, pid):
        RtNetlink.__init__(self, pid)
        self.logger = logging.getLogger('ifupdown.' +
                            self.__class__.__name__)
        self.bind(0, None)
        self.bind_done = True
        self.ifindexmap = {}

    def do_bind(self):
        if self.bind_done:
            return True
        self.bind(0, None)
        self.bind_done = True

    def get_ifindex(self, ifname):
        ifindex = self.ifindexmap.get(ifname)
        if not ifindex:
            with open('/sys/class/net/%s/ifindex' %ifname, 'r') as f:
                ifindex = int(f.read())
                self.ifindexmap[ifname] = ifindex
        return ifindex

    def create_vlan(self, link, ifname, vlanid):
        self.logger.info('rtnetlink: creating vlan interface %s' %ifname)
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return
        try:
            ifindex = self.get_ifindex(link)
        except Exception as e:
            raise Exception('cannot determine ifindex for link %s (%s)'
                            %(link, str(e)))

        ifm = Ifinfomsg(AF_UNSPEC)
        rtas = {IFLA_IFNAME: ifname,
                    IFLA_LINK : ifindex,
		            IFLA_LINKINFO : {
			            IFLA_INFO_KIND : 'vlan',
			            IFLA_INFO_DATA : {
                            IFLA_VLAN_ID : vlanid,
                        }
                    }
               }
        token = self.request(RTM_NEWLINK,
                        NLM_F_CREATE | NLM_F_REQUEST | NLM_F_ACK, ifm, rtas)
        self.process_wait([token])

    def create_macvlan(self, ifname, link, mode='private'):
        self.logger.info('rtnetlink: creating macvlan interface %s' %ifname)
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return
        try:
            ifindex = self.get_ifindex(link)
        except Exception as e:
            raise Exception('cannot determine ifindex for link %s (%s)'
                            %(link, str(e)))

        ifm = Ifinfomsg(AF_UNSPEC)
        rtas = {IFLA_IFNAME: ifname,
                    IFLA_LINK : ifindex,
		            IFLA_LINKINFO : {
			            IFLA_INFO_KIND : 'macvlan',
		            IFLA_INFO_DATA : {
                           IFLA_MACVLAN_MODE : MACVLAN_MODE_PRIVATE,
                        }
                    }
               }
        token = self.request(RTM_NEWLINK, NLM_F_CREATE | NLM_F_REQUEST |
                             NLM_F_ACK, ifm, rtas)
        self.process_wait([token])

    def link_set(self, ifname, state):
        flags = 0
        self.logger.info('rtnetlink: setting link %s %s' %(ifname, state))
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return

        if state == "up":
            flags |= IFF_UP
        else:
            flags &= ~IFF_UP

        ifm = Ifinfomsg(AF_UNSPEC, ifi_change=IFF_UP, ifi_flags=flags)
        rtas = {IFLA_IFNAME: ifname}

        token = self.request(RTM_NEWLINK, NLM_F_REQUEST | NLM_F_ACK, ifm, rtas)
        self.process_wait([token])

    def link_set_hwaddress(self, ifname, hwaddress):
        flags = 0
        self.logger.info('rtnetlink: setting link hwaddress %s %s' %(ifname, hwaddress))
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return

        flags &= ~IFF_UP
        ifm = Ifinfomsg(AF_UNSPEC, ifi_change=IFF_UP)
        rtas = {IFLA_IFNAME: ifname,
                IFLA_ADDRESS : str(bytearray([int(a,16) for a in hwaddress.split(':')]))}

        self.logger.info(rtas)

        token = self.request(RTM_NEWLINK, NLM_F_REQUEST | NLM_F_ACK, ifm, rtas)
        self.process_wait([token])

    def addr_add(self, ifname, address, broadcast=None, peer=None, scope=None,
                 preferred_lifetime=None):
        self.logger.info('rtnetlink: setting address')
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return

        try:
            ifindex = self.get_ifindex(link)
        except Exception as e:
            raise Exception('cannot determine ifindex for link %s (%s)'
                            %(link, str(e)))
        ifa_scope = RT_SCOPE_
        if scope:
            if scope == "universe":
                ifa_scope = RT_SCOPE_UNIVERSE
            elif scope == "site":
                ifa_scope = RT_SCOPE_SITE
            elif scope == "link":
                ifa_scope = RT_SCOPE_LINK
            elif scope == "host":
                ifa_scope = RT_SCOPE_HOST
            elif scope == "nowhere":
                ifa_scope = RT_SCOPE_NOWHERE
        rtas = {IFLA_ADDRESS: ifname}

        ifa = Ifaddrmsg(AF_UNSPEC, ifa_scope=ifa_scope, ifa_index=ifindex)

        token = self.request(RTM_NEWADDR, NLM_F_REQUEST | NLM_F_ACK, ifa, rtas)
        self.process_wait([token])

    def link_set_many(self, ifname, ifattrs):
        _ifattr_to_rta_map = {'dev' : IFLA_NAME,
                              'address' : IFLA_ADDRESS,
                              'broadcast' : IFLA_BROADCAST,
                              'mtu' : IFLA_MTU,
                              'master' : IFLA_MASTER}
        flags = 0
        ifi_change = IFF_UP
        rtas = {}
        self.logger.info('rtnetlink: setting link %s %s' %(ifname, state))
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return
        if not ifattrs:
           return
        state = ifattrs.get('state')
        if state == 'up':
            flags |= IFF_UP
        elif state == 'down':
            flags &= ~IFF_UP
        else:
            ifi_change = 0

        if ifi_change:
           ifm = Ifinfomsg(AF_UNSPEC, ifi_change=IFF_UP, ifi_flags=flags)
        else:
           ifm = Ifinfomsg(AF_UNSPEC)

        for attr, attrval in ifattrs.items():
            rta_attr = _ifattr_to_rta_map.get(attr)
            if rta_attr:
               if attr == 'hwaddress':
                  rtas[rta_attr] = str(bytearray([int(a,16) for a in attrval.split(':')]))
               else:
                  rtas[rta_attr] = attrval

        token = self.request(RTM_NEWLINK, NLM_F_REQUEST | NLM_F_ACK, ifm, rtas)
        self.process_wait([token])

    def bridge_vlan(self, add=True, vid=None, dev=None, pvid=False,
                    untagged=False, master=True):
        flags = 0
        vflags = 0
        if not vid or not dev:
           return
        self.logger.info('rtnetlink: bridge vlan add vid %s %s %s dev %s %s'
                         %(vid, 'untagged' if untagged else '',
                           'pvid' if pvid else '', dev,
                           'self' if self else ''))
        if ifupdownmain.ifupdownFlags.DRYRUN:
            return
        try:
            ifindex = self.get_ifindex(dev)
        except Exception as e:
            raise Exception('cannot determine ifindex for dev %s (%s)'
                            %(dev, str(e)))
        if not master:
            flags = BRIDGE_FLAGS_SELF

        if pvid:
           vflags = BRIDGE_VLAN_INFO_PVID
           vflags |= BRIDGE_VLAN_INFO_UNTAGGED
        elif untagged:
           vflags |= BRIDGE_VLAN_INFO_UNTAGGED

        ifm = Ifinfomsg(AF_BRIDGE, ifi_index=ifindex)
        rtas = {IFLA_AF_SPEC: {
                    IFLA_BRIDGE_FLAGS: flags,
                    IFLA_BRIDGE_VLAN_INFO : BridgeVlanInfo(vflags, int(vid), int(vid))
                  }
               }
        if add:
            token = self.request(RTM_SETLINK,
                        NLM_F_REQUEST | NLM_F_ACK, ifm, rtas)
        else:
            token = self.request(RTM_DELLINK,
                        NLM_F_REQUEST | NLM_F_ACK, ifm, rtas)
        self.process_wait([token])

    def bridge_vlan_many(self, add=True, vids=[], dev=None, pvid=False,
                         untagged=False, master=True):
        for v in vids:
            self.bridge_vlan_add(add, v, dev, ispvid, isuntagged, master)

rtnl_api = rtnetlinkApi(os.getpid())



#!/usr/bin/python
#
# Copyright 2014 Cumulus Networks, Inc. All rights reserved.
# Author: Roopa Prabhu, roopa@cumulusnetworks.com
#

from sets import Set
from ifupdown.iface import *
import ifupdownaddons
from ifupdownaddons.modulebase import moduleBase
from ifupdownaddons.ifenslaveutil import ifenslaveutil
from ifupdownaddons.iproute2 import iproute2
import ifupdown.rtnetlink_api as rtnetlink_api

class ifenslave(moduleBase):
    """  ifupdown2 addon module to configure bond interfaces """
    _modinfo = { 'mhelp' : 'bond configuration module',
                    'attrs' : {
                    'bond-use-carrier':
                         {'help' : 'bond use carrier',
                          'validvals' : ['0', '1'],
                          'default' : '1',
                          'example': ['bond-use-carrier 1']},
                     'bond-num-grat-arp':
                         {'help' : 'bond use carrier',
                          'validrange' : ['0', '255'],
                          'default' : '1',
                          'example' : ['bond-num-grat-arp 1']},
                     'bond-num-unsol-na' :
                         {'help' : 'bond slave devices',
                          'validrange' : ['0', '255'],
                          'default' : '1',
                          'example' : ['bond-num-unsol-na 1']},
                     'bond-xmit-hash-policy' :
                         {'help' : 'bond slave devices',
                          'validvals' : ['layer2', 'layer3+4', 'layer2+3'],
                          'default' : 'layer2',
                          'example' : ['bond-xmit-hash-policy layer2']},
                     'bond-miimon' :
                         {'help' : 'bond miimon',
                          'validrange' : ['0', '255'],
                          'default' : '0',
                          'example' : ['bond-miimon 0']},
                     'bond-mode' :
                         {'help' : 'bond mode',
                          'validvals' : ['balance-rr', 'active-backup',
                                          'balance-xor', 'broadcast', '802.3ad',
                                          'balance-tlb', 'balance-alb'],
                          'default' : 'balance-rr',
                          'example' : ['bond-mode 802.3ad']},
                     'bond-lacp-rate':
                         {'help' : 'bond lacp rate',
                          'validvals' : ['0', '1'],
                          'default' : '0',
                          'example' : ['bond-lacp-rate 0']},
                     'bond-min-links':
                         {'help' : 'bond min links',
                          'default' : '0',
                          'example' : ['bond-min-links 0']},
                     'bond-ad-sys-priority':
                         {'help' : '802.3ad system priority',
                          'default' : '65535',
                          'example' : ['bond-ad-sys-priority 65535']},
                     'bond-ad-sys-mac-addr':
                         {'help' : '802.3ad system mac address',
                          'default' : '00:00:00:00:00:00',
                         'example' : ['bond-ad-sys-mac-addr 00:00:00:00:00:00']},
                     'bond-lacp-fallback-allow':
                         {'help' : 'allow lacp fall back',
                          'compat' : True,
                          'validvals' : ['0', '1'],
                          'default' : '0',
                          'example' : ['bond-lacp-fallback-allow 0']},
                     'bond-lacp-fallback-period':
                         {'help' : 'grace period (seconds) for lacp fall back',
                          'compat' : True,
                          'validrange' : ['0', '100'],
                          'default' : '90',
                          'example' : ['bond-lacp-fallback-period 100']},
                     'bond-lacp-fallback-priority':
                         {'help' : 'slave priority for lacp fall back',
                          'compat' : True,
                          'example' : ['bond-lacp-fallback-priority swp1=1 swp2=1 swp3=2']},
                     'bond-lacp-bypass-allow':
                         {'help' : 'allow lacp bypass',
                          'validvals' : ['0', '1'],
                          'default' : '0',
                          'example' : ['bond-lacp-bypass-allow 0']},
                     'bond-lacp-bypass-period':
                         {'help' : 'grace period (seconds) for lacp bypass',
                          'validrange' : ['0', '900'],
                          'default' : '0',
                          'example' : ['bond-lacp-bypass-period 100']},
                     'bond-lacp-bypass-priority':
                         {'help' : 'slave priority for lacp bypass',
                          'example' : ['bond-lacp-bypass-priority swp1=1 swp2=1 swp3=2']},
                     'bond-slaves' :
                        {'help' : 'bond slaves',
                         'required' : True,
                         'example' : ['bond-slaves swp1 swp2',
                                      'bond-slaves glob swp1-2',
                                      'bond-slaves regex (swp[1|2)']}}}

    def __init__(self, *args, **kargs):
        ifupdownaddons.modulebase.moduleBase.__init__(self, *args, **kargs)
        self.ipcmd = None
        self.ifenslavecmd = None

    def _is_bond(self, ifaceobj):
        if ifaceobj.get_attr_value_first('bond-slaves'):
            return True
        return False

    def get_dependent_ifacenames(self, ifaceobj, ifacenames_all=None):
        """ Returns list of interfaces dependent on ifaceobj """

        if not self._is_bond(ifaceobj):
            return None
        slave_list = self.parse_port_list(ifaceobj.get_attr_value_first(
                                    'bond-slaves'), ifacenames_all)
        ifaceobj.dependency_type = ifaceDependencyType.MASTER_SLAVE
        # Also save a copy for future use
        ifaceobj.priv_data = list(slave_list)
        if ifaceobj.link_type != ifaceLinkType.LINK_NA:
           ifaceobj.link_type = ifaceLinkType.LINK_MASTER
        ifaceobj.link_kind |= ifaceLinkKind.BOND
        ifaceobj.role |= ifaceRole.MASTER

        return slave_list

    def get_dependent_ifacenames_running(self, ifaceobj):
        self._init_command_handlers()
        return self.ifenslavecmd.get_slaves(ifaceobj.name)

    def _get_slave_list(self, ifaceobj):
        """ Returns slave list present in ifaceobj config """

        # If priv data already has slave list use that first.
        if ifaceobj.priv_data:
            return ifaceobj.priv_data
        slaves = ifaceobj.get_attr_value_first('bond-slaves')
        if slaves:
            return self.parse_port_list(slaves)
        else:
            return None

    def fetch_attr(self, ifaceobj, attrname):
        attrval = ifaceobj.get_attr_value_first(attrname)
        if attrval:
            msg = ('%s: invalid value %s for attr %s.'
                    %(ifaceobj.name, attrval, attrname))
            optiondict = self.get_mod_attr(attrname)
            if not optiondict:
                return None
            validvals = optiondict.get('validvals')
            if validvals and attrval not in validvals:
                raise Exception(msg + ' Valid values are %s' %str(validvals))
            validrange = optiondict.get('validrange')
            if validrange:
                if (int(attrval) < int(validrange[0]) or
                        int(attrval) > int(validrange[1])):
                    raise Exception(msg + ' Valid range is [%s,%s]'
                                    %(validrange[0], validrange[1]))
            if attrname == 'bond-mode' and attrval == '802.3ad':
               dattrname = 'bond-min-links'
               min_links = ifaceobj.get_attr_value_first(dattrname)
               if not min_links or min_links == '0':
                   self.logger.warn('%s: required attribute %s'
                        %(ifaceobj.name, dattrname) +
                        ' not present or set to \'0\'')
        return attrval

    def _apply_master_settings(self, ifaceobj):
        have_attrs_to_set = 0
        linkup = False
        ifenslavecmd_attrmap =  OrderedDict([('bond-mode' , 'mode'),
                                 ('bond-miimon' , 'miimon'),
                                 ('bond-use-carrier', 'use_carrier'),
                                 ('bond-lacp-rate' , 'lacp_rate'),
                                 ('bond-xmit-hash-policy' , 'xmit_hash_policy'),
                                 ('bond-min-links' , 'min_links'),
                                 ('bond-num-grat-arp' , 'num_grat_arp'),
                                 ('bond-num-unsol-na' , 'num_unsol_na'),
                                 ('bond-ad-sys-mac-addr' , 'ad_sys_mac_addr'),
                                 ('bond-ad-sys-priority' , 'ad_sys_priority'),
                                 ('bond-lacp-fallback-allow', 'lacp_bypass_allow'),
                                 ('bond-lacp-fallback-period', 'lacp_bypass_period'),
                                 ('bond-lacp-bypass-allow', 'lacp_bypass_allow'),
                                 ('bond-lacp-bypass-period', 'lacp_bypass_period')])
        linkup = self.ipcmd.is_link_up(ifaceobj.name)
        try:
            # order of attributes set matters for bond, so
            # construct the list sequentially
            attrstoset = OrderedDict()
            for k, dstk in ifenslavecmd_attrmap.items():
                v = self.fetch_attr(ifaceobj, k)
                if v:
                    attrstoset[dstk] = v
            if not attrstoset:
                return
            have_attrs_to_set = 1
            self.ifenslavecmd.set_attrs(ifaceobj.name, attrstoset,
                    self.ipcmd.link_down if linkup else None)
        except:
            raise
        finally:
            if have_attrs_to_set and linkup:
                self.ipcmd.link_up(ifaceobj.name)

    def _add_slaves(self, ifaceobj):
        runningslaves = []

        slaves = self._get_slave_list(ifaceobj)
        if not slaves:
            self.logger.debug('%s: no slaves found' %ifaceobj.name)
            return

        if not self.PERFMODE:
            runningslaves = self.ifenslavecmd.get_slaves(ifaceobj.name);

        for slave in Set(slaves).difference(Set(runningslaves)):
            if not self.PERFMODE and not self.ipcmd.link_exists(slave):
                    self.log_warn('%s: skipping slave %s, does not exist'
                                  %(ifaceobj.name, slave))
                    continue
            link_up = False
            if self.ipcmd.is_link_up(slave):
               rtnetlink_api.rtnl_api.link_set(slave, "down")
               link_up = True
            self.ipcmd.link_set(slave, 'master', ifaceobj.name)
            if link_up or ifaceobj.link_type != ifaceLinkType.LINK_NA:
               try:
                    rtnetlink_api.rtnl_api.link_set(slave, "up")
               except Exception as e:
                    self.logger.debug('%s: %s: link set up (%s)'
                                      %(ifaceobj.name, slave, str(e)))
                    pass

        if runningslaves:
            # Delete active slaves not in the new slave list
            [ self.ifenslavecmd.remove_slave(ifaceobj.name, s)
                    for s in runningslaves if s not in slaves ]

    def _apply_slaves_lacp_bypass_prio(self, ifaceobj):
        slaves = self.ifenslavecmd.get_slaves(ifaceobj.name)
        if not slaves:
           return
        attrval = ifaceobj.get_attrs_value_first(['bond-lacp-bypass-priority',
                                'bond-lacp-fallback-priority'])
        if attrval:
            portlist = self.parse_port_list(attrval)
            if not portlist:
                self.log_warn('%s: could not parse \'%s %s\''
                              %(ifaceobj.name, attrname, attrval))
                return
            for p in portlist:
                try:
                    (port, val) = p.split('=')
                    if port not in slaves:
                        self.log_warn('%s: skipping slave %s, does not exist' 
                                      %(ifaceobj.name, port))
                        continue
                    slaves.remove(port)
                    self.ifenslavecmd.set_lacp_fallback_priority(
                                            ifaceobj.name, port, val)
                except Exception as e:
                    self.log_warn('%s: failed to set lacp_fallback_priority %s (%s)'
                                  %(ifaceobj.name, port, str(e)))

        for p in slaves:
            try:
                self.ifenslavecmd.set_lacp_fallback_priority(ifaceobj.name, p, '0')
            except Exception as e:
                self.log_warn('%s: failed to clear lacp_bypass_priority %s (%s)'
                              %(ifaceobj.name, p, str(e)))


    def _up(self, ifaceobj):
        try:
            if not self.ipcmd.link_exists(ifaceobj.name):
                self.ifenslavecmd.create_bond(ifaceobj.name)
            self._apply_master_settings(ifaceobj)
            self._add_slaves(ifaceobj)
            self._apply_slaves_lacp_bypass_prio(ifaceobj)
            if ifaceobj.addr_method == 'manual':
               rtnetlink_api.rtnl_api.link_set(ifaceobj.name, "up")
        except Exception as e:
            self.log_error(str(e))

    def _down(self, ifaceobj):
        try:
            self.ifenslavecmd.delete_bond(ifaceobj.name)
        except Exception as e:
            self.log_warn(str(e))

    def _query_check(self, ifaceobj, ifaceobjcurr):
        slaves = None

        if not self.ifenslavecmd.bond_exists(ifaceobj.name):
            self.logger.debug('bond iface %s' %ifaceobj.name +
                              ' does not exist')
            return

        ifaceattrs = self.dict_key_subset(ifaceobj.config,
                                          self.get_mod_attrs())
        if not ifaceattrs: return
        runningattrs = self._query_running_attrs(ifaceobj.name)

        # backward compat change
        runningattrs.update({'bond-lacp-fallback-allow': runningattrs.get(
                                                    'bond-lacp-bypass-allow'),
                          'bond-lacp-fallback-period': runningattrs.get(
                                                    'bond-lacp-bypass-period'),
                          'bond-lacp-fallback-priority': runningattrs.get(
                                                'bond-lacp-bypass-priority')})
        for k in ifaceattrs:
            v = ifaceobj.get_attr_value_first(k)
            if not v:
                continue
            if k == 'bond-slaves':
                slaves = self._get_slave_list(ifaceobj)
                continue
            rv = runningattrs.get(k)
            if not rv:
                ifaceobjcurr.update_config_with_status(k, 'None', 1)
            else:
                if (k == 'bond-lacp-bypass-priority' or
                    k == 'bond-lacp-fallback-priority'):
                    prios = v.split()
                    prios.sort()
                    prio_str = ' '.join(prios)
                    ifaceobjcurr.update_config_with_status(k, rv,
                                    1 if prio_str != rv else 0)
                    continue
                ifaceobjcurr.update_config_with_status(k, rv,
                                                       1 if v != rv else 0)
        runningslaves = runningattrs.get('bond-slaves')
        if not slaves and not runningslaves:
            return
        retslave = 1
        if slaves and runningslaves:
            if slaves and runningslaves:
                difference = set(slaves).symmetric_difference(runningslaves)
                if not difference:
                    retslave = 0
        ifaceobjcurr.update_config_with_status('bond-slaves',
                        ' '.join(runningslaves)
                        if runningslaves else 'None', retslave)

    def _query_running_attrs(self, bondname):
        bondattrs = {'bond-mode' :
                            self.ifenslavecmd.get_mode(bondname),
                     'bond-miimon' :
                            self.ifenslavecmd.get_miimon(bondname),
                     'bond-use-carrier' :
                            self.ifenslavecmd.get_use_carrier(bondname),
                     'bond-lacp-rate' :
                            self.ifenslavecmd.get_lacp_rate(bondname),
                     'bond-min-links' :
                            self.ifenslavecmd.get_min_links(bondname),
                     'bond-ad-sys-mac-addr' :
                            self.ifenslavecmd.get_ad_sys_mac_addr(bondname),
                     'bond-ad-sys-priority' :
                            self.ifenslavecmd.get_ad_sys_priority(bondname),
                     'bond-xmit-hash-policy' :
                            self.ifenslavecmd.get_xmit_hash_policy(bondname),
                     'bond-lacp-bypass-allow' :
                            self.ifenslavecmd.get_lacp_fallback_allow(bondname),
                     'bond-lacp-bypass-period' :
                            self.ifenslavecmd.get_lacp_fallback_period(bondname),
                     'bond-lacp-bypass-priority' :
                            self.ifenslavecmd.get_lacp_fallback_priority(bondname)}
        slaves = self.ifenslavecmd.get_slaves(bondname)
        if slaves:
            bondattrs['bond-slaves'] = slaves
        return bondattrs

    def _query_running(self, ifaceobjrunning):
        if not self.ifenslavecmd.bond_exists(ifaceobjrunning.name):
            return
        bondattrs = self._query_running_attrs(ifaceobjrunning.name)
        if bondattrs.get('bond-slaves'):
            bondattrs['bond-slaves'] = ' '.join(bondattrs.get('bond-slaves'))
        [ifaceobjrunning.update_config(k, v)
                    for k, v in bondattrs.items()
                        if v and v != self.get_mod_subattr(k, 'default')]

    _run_ops = {'pre-up' : _up,
               'post-down' : _down,
               'query-running' : _query_running,
               'query-checkcurr' : _query_check}

    def get_ops(self):
        """ returns list of ops supported by this module """
        return self._run_ops.keys()

    def _init_command_handlers(self):
        flags = self.get_flags()
        if not self.ipcmd:
            self.ipcmd = iproute2(**flags)
        if not self.ifenslavecmd:
            self.ifenslavecmd = ifenslaveutil(**flags)

    def run(self, ifaceobj, operation, query_ifaceobj=None, **extra_args):
        """ run bond configuration on the interface object passed as argument

        Args:
            **ifaceobj** (object): iface object

            **operation** (str): any of 'pre-up', 'post-down', 'query-checkcurr',
                'query-running'

        Kwargs:
            **query_ifaceobj** (object): query check ifaceobject. This is only
                valid when op is 'query-checkcurr'. It is an object same as
                ifaceobj, but contains running attribute values and its config
                status. The modules can use it to return queried running state
                of interfaces. status is success if the running state is same
                as user required state in ifaceobj. error otherwise.
        """
        op_handler = self._run_ops.get(operation)
        if not op_handler:
            return
        if operation != 'query-running' and not self._is_bond(ifaceobj):
            return
        self._init_command_handlers()
        if operation == 'query-checkcurr':
            op_handler(self, ifaceobj, query_ifaceobj)
        else:
            op_handler(self, ifaceobj)
